# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Dict
from typing import Optional
from typing import Union
from unittest import TestCase

import pandas as pd
from parameterized.parameterized import parameterized

from kats.compat.pandas import assert_frame_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_data  # @manual
from kats.models.arima import ARIMAModel
from kats.models.arima import ARIMAParams
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_15_ARIMA_PARAM_1_MODEL_1,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_15_ARIMA_PARAM_1_MODEL_1_INCL_HIST,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_15_ARIMA_PARAM_1_MODEL_2,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_15_ARIMA_PARAM_2_MODEL_1,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_15_ARIMA_PARAM_2_MODEL_2,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_ARIMA_PARAM_1_MODEL_1,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_ARIMA_PARAM_1_MODEL_1_INCL_HIST,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_ARIMA_PARAM_1_MODEL_2,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_ARIMA_PARAM_2_MODEL_1,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_ARIMA_PARAM_2_MODEL_2,
)

STEPS_1 = 15
STEPS_2 = 30
# pyre-fixme[5]: Global expression must be annotated.
TEST_DATA = {
    "daily": {
        "ts": TimeSeriesData(load_data("peyton_manning.csv")),
        "invalid_ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv")),
        # "ts_nan": TimeSeriesData(df=PEYTON_INPUT_NAN),
        "freq": "D",
        "p1": ARIMAParams(p=1, d=1, q=1),
        "p2": ARIMAParams(p=1, d=1, q=3),
        "invalid_p": ARIMAParams(p=1, d=2, q=1),
        "m1": {
            "method": "css-mle",
            "trend": "c",
            "solver": "lbfgs",
            "maxiter": 500,
            "full_output": False,
            "disp": 0,
            "start_ar_lags": None,
        },
        "m2": {
            "method": "css",
            "trend": "nc",
            "solver": "newton",
            "maxiter": 2,
            "full_output": False,
            "disp": 0,
            "start_ar_lags": 10,
        },
        "invalid_m1": {
            "method": "css-mle",
            "trend": "c",
            "solver": "lbfgs",
            "maxiter": 500,
            "full_output": False,
            "disp": 0,
            "start_ar_lags": 2,
        },
        "invalid_m2": {
            "method": "css-mle",
            "trend": "c",
            "solver": "invalid",
            "maxiter": 500,
            "full_output": False,
            "disp": 0,
            "start_ar_lags": None,
        },
        "invalid_m3": {
            "method": "css-mle",
            "trend": "invalid",
            "solver": "lbfgs",
            "maxiter": 500,
            "full_output": False,
            "disp": 0,
            "start_ar_lags": None,
        },
        "invalid_m4": {
            "method": "invalid",
            "trend": "c",
            "solver": "lbfgs",
            "maxiter": 500,
            "full_output": False,
            "disp": 0,
            "start_ar_lags": None,
        },
        "steps_1": STEPS_1,
        "steps_2": STEPS_2,
        "no_incl_hist": False,
        "incl_hist": True,
        "truth_p1_m1_15": PEYTON_FCST_15_ARIMA_PARAM_1_MODEL_1,
        "truth_p2_m1_15": PEYTON_FCST_15_ARIMA_PARAM_2_MODEL_1,
        "truth_p1_m2_15": PEYTON_FCST_15_ARIMA_PARAM_1_MODEL_2,
        "truth_p2_m2_15": PEYTON_FCST_15_ARIMA_PARAM_2_MODEL_2,
        "truth_p1_m1_30": PEYTON_FCST_30_ARIMA_PARAM_1_MODEL_1,
        "truth_p2_m1_30": PEYTON_FCST_30_ARIMA_PARAM_2_MODEL_1,
        "truth_p1_m2_30": PEYTON_FCST_30_ARIMA_PARAM_1_MODEL_2,
        "truth_p2_m2_30": PEYTON_FCST_30_ARIMA_PARAM_2_MODEL_2,
        "truth_p1_m1_15_incl_hist": PEYTON_FCST_15_ARIMA_PARAM_1_MODEL_1_INCL_HIST,
        "truth_p1_m1_30_incl_hist": PEYTON_FCST_30_ARIMA_PARAM_1_MODEL_1_INCL_HIST,
    },
}


class ARIMAModelTest(TestCase):
    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "daily_p1_m1",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["m1"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p1_m1_15"],
                TEST_DATA["daily"]["truth_p1_m1_30"],
                TEST_DATA["daily"]["no_incl_hist"],
            ],
            [
                "daily_p2_m1",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p2"],
                TEST_DATA["daily"]["m1"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p2_m1_15"],
                TEST_DATA["daily"]["truth_p2_m1_30"],
                TEST_DATA["daily"]["no_incl_hist"],
            ],
            [
                "daily_p1_m2",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["m2"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p1_m2_15"],
                TEST_DATA["daily"]["truth_p1_m2_30"],
                TEST_DATA["daily"]["no_incl_hist"],
            ],
            [
                "daily_p2_m2",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p2"],
                TEST_DATA["daily"]["m2"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p2_m2_15"],
                TEST_DATA["daily"]["truth_p2_m2_30"],
                TEST_DATA["daily"]["no_incl_hist"],
            ],
            [
                "daily_p1_m1_incl_hist",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["m1"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p1_m1_15_incl_hist"],
                TEST_DATA["daily"]["truth_p1_m1_30_incl_hist"],
                TEST_DATA["daily"]["incl_hist"],
            ],
        ]
    )
    def test_fcst(
        self,
        name: str,
        ts: TimeSeriesData,
        params: ARIMAParams,
        model_params: Dict[str, Optional[Union[str, int, bool]]],
        steps_1: int,
        steps_2: int,
        truth_1: pd.DataFrame,
        truth_2: pd.DataFrame,
        include_history: bool = False,
    ) -> None:
        m = ARIMAModel(data=ts, params=params)
        # pyre-fixme[6]: Incompatible parameter type...
        m.fit(**model_params)
        res_1 = m.predict(steps=steps_1, include_history=include_history)
        res_2 = m.predict(steps=steps_2, include_history=include_history)
        if include_history:
            assert_frame_equal(
                res_1.reset_index(drop=True),
                truth_1.reset_index(drop=True),
                check_exact=False,
                check_less_precise=True,
                rtol=0.001,
            )
            assert_frame_equal(
                res_2.reset_index(drop=True),
                truth_2.reset_index(drop=True),
                check_exact=False,
                check_less_precise=True,
                rtol=0.001,
            )
        else:
            assert_frame_equal(
                res_1,
                truth_1,
                check_less_precise=True,
                check_exact=False,
                rtol=0.001,
            )
            assert_frame_equal(
                res_2,
                truth_2,
                check_less_precise=True,
                check_exact=False,
                rtol=0.001,
            )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "invalid_p",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["invalid_p"],
                TEST_DATA["daily"]["m1"],
            ],
            [
                "invalid_m1",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["invalid_m1"],
            ],
            [
                "invalid_m2",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["invalid_m2"],
            ],
            [
                "invalid_m3",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["invalid_m3"],
            ],
            [
                "invalid_m4",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["invalid_m4"],
            ],
            [
                "invalid_ts",
                TEST_DATA["daily"]["invalid_ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["m1"],
            ],
        ]
    )
    def test_invalid_params(
        self,
        name: str,
        ts: TimeSeriesData,
        params: ARIMAParams,
        model_params: Dict[str, Optional[Union[str, int, bool]]],
    ) -> None:
        with self.assertRaises(ValueError):
            m = ARIMAModel(data=ts, params=params)
            # pyre-fixme[6]: Incompatible parameter type...
            m.fit(**model_params)

    def test_exec_plot(self) -> None:
        m = ARIMAModel(data=TEST_DATA["daily"]["ts"], params=TEST_DATA["daily"]["p1"])
        m.fit(**TEST_DATA["daily"]["m1"])
        _ = m.predict(steps=STEPS_1)
        m.plot()

    def test_name(self) -> None:
        m = ARIMAModel(data=TEST_DATA["daily"]["ts"], params=TEST_DATA["daily"]["p1"])
        self.assertEqual(m.__str__(), "ARIMA")

    def test_search_space(self) -> None:
        m = ARIMAModel(data=TEST_DATA["daily"]["ts"], params=TEST_DATA["daily"]["p1"])
        self.assertEqual(
            m.get_parameter_search_space(),
            [
                {
                    "name": "p",
                    "type": "choice",
                    "values": list(range(1, 6)),
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "d",
                    "type": "choice",
                    "values": list(range(1, 3)),
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "q",
                    "type": "choice",
                    "values": list(range(1, 6)),
                    "value_type": "int",
                    "is_ordered": True,
                },
            ],
        )

    # Filler test
    def test_validate_params(self) -> None:
        params = TEST_DATA["daily"]["p1"]
        params.validate_params()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""A module for meta-learner predictability.

This module contains the class :class:`MetaLearnPredictability` for meta-learner predictability. This class predicts whether a time series is predictable or not.
The predictability of a time series is determined by whether the forecasting errors of the possible best forecasting model can be less than a user-defined threshold.
"""

import ast
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import joblib
import numpy as np
import pandas as pd
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import precision_recall_fscore_support
from sklearn.model_selection import train_test_split
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier

from kats.consts import TimeSeriesData
from kats.tsfeatures.tsfeatures import TsFeatures


class MetaLearnPredictability:
    """Meta-learner framework on predictability.
    This framework uses classification algorithms to predict whether a time series is predictable or not (
    we define the time series with error metrics less than a user defined threshold as predictable).
    For training, it uses time series features as inputs and whether the best forecasting models' errors less than the user-defined threshold as labels.
    For prediction, it takes time series or time series features as inputs to predict whether the corresponding time series is predictable or not.
    This class provides preprocess, pred, pred_by_feature, save_model and load_model.

    Attributes:
        metadata: Optional; A list of dictionaries representing the meta-data of time series.
                  Each dictionary d must contain at least 3 components: 'hpt_res', 'features' and 'best_model'. d['hpt_res'] represents the best hyper-parameters for each candidate model and the corresponding errors;
                  d['features'] are time series features, and d['best_model'] is a string representing the best candidate model of the corresponding time series data.
                  metadata should not be None unless load_model is True. Default is None
        threshold: Optional; A float representing the threshold for the forecasting error. A time series whose forecasting error of the best forecasting model is higher than the threshold is considered as unpredictable. Default is 0.2.
        load_model: Optional; A boolean to specify whether or not to load a trained model. Default is False.

    Sample Usage:
        >>> mlp = MetaLearnPredictability(data)
        >>> mlp.train()
        >>> mlp.save_model()
        >>> mlp.pred(TSdata) # Predict whether a time series is predictable.
        >>> mlp2 = MetaLearnPredictability(load_model=True) # Create a new object to load the trained model
        >>> mlp2.load_model()
    """

    def __init__(
        self,
        # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
        metadata: Optional[List[Any]] = None,
        threshold: float = 0.2,
        # pyre-fixme[2]: Parameter must be annotated.
        load_model=False,
    ) -> None:
        if load_model:
            msg = "Initialize this class without meta data, and a pretrained model should be loaded using .load_model() method."
            logging.info(msg)
        else:
            if metadata is None:
                msg = "Please input meta data to initialize this class."
                logging.error(msg)
                raise ValueError(msg)
            if len(metadata) <= 30:
                msg = "Dataset is too small to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            if "hpt_res" not in metadata[0]:
                msg = "Missing best hyper-params, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            if "features" not in metadata[0]:
                msg = "Missing time series features, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            if "best_model" not in metadata[0]:
                msg = "Missing best models, not able to train a meta learner!"
                logging.error(msg)
                raise ValueError(msg)

            # pyre-fixme[4]: Attribute must be annotated.
            self.metadata = metadata
            # pyre-fixme[4]: Attribute must be annotated.
            self.threshold = threshold
            self._reorganize_data()
            self._validate_data()
            # pyre-fixme[4]: Attribute must be annotated.
            self.rescale = False
            # pyre-fixme[4]: Attribute must be annotated.
            self.clf = None
            # pyre-fixme[4]: Attribute must be annotated.
            self._clf_threshold = None

    def _reorganize_data(self) -> None:
        """Reorganize raw input data into features and labels."""

        metadata = self.metadata

        self.features = []
        self.labels = []

        for i in range(len(metadata)):
            try:
                if isinstance(metadata[i]["hpt_res"], str):
                    hpt = ast.literal_eval(metadata[i]["hpt_res"])
                else:
                    hpt = metadata[i]["hpt_res"]

                if isinstance(metadata[i]["features"], str):
                    feature = ast.literal_eval(metadata[i]["features"])
                else:
                    feature = metadata[i]["features"]

                self.features.append(feature)
                self.labels.append(hpt[metadata[i]["best_model"]][1])
            except Exception as e:
                logging.exception(e)
        self.labels = (np.array(self.labels) > self.threshold).astype(int)
        self.features = pd.DataFrame(self.features, copy=False)
        self.features.fillna(0, inplace=True)
        self.features_mean = np.average(self.features.values, axis=0)

        self.features_std = np.std(self.features.values, axis=0)

        self.features_std[self.features_std == 0] = 1.0

        return

    def _validate_data(self) -> None:
        """Validate input data.

        We check two aspects:
            1) whether input data contain both positive and negative instances.
            2) whether training data size is at least 30.
        """

        if len(np.unique(self.labels)) == 1:
            msg = "Only one type of time series data and cannot train a classifier!"
            logging.error(msg)
            raise ValueError(msg)
        if len(self.features) <= 30:
            msg = "Dataset is too small to train a meta learner!"
            logging.error(msg)
            raise ValueError(msg)

    def preprocess(self) -> None:
        """Rescale input time series features to zero-mean and unit-variance.

        Returns:
            None.
        """

        self.rescale = True
        features = (self.features.values - self.features_mean) / self.features_std
        self.features = pd.DataFrame(features, columns=self.features.columns, copy=False)

    def train(
        self,
        method: str = "RandomForest",
        valid_size: float = 0.1,
        test_size: float = 0.1,
        recall_threshold: float = 0.7,
        n_estimators: int = 500,
        n_neighbors: int = 5,
        # pyre-fixme[2]: Parameter must be annotated.
        **kwargs,
    ) -> Dict[str, float]:
        """Train a classifier with time series features to forecast predictability.

        Args:
            method: Optional; A string representing the name of the classification algorithm. Can be 'RandomForest', 'GBDT', 'KNN' or 'NaiveBayes'. Default is 'RandomForest'.
            valid_size: Optional; A float representing the size of validation set for parameter tunning, which should be within (0, 1). Default is 0.1.
            test_size: Optional; A float representing the size of test set, which should be within [0., 1-valid_size). Default is 0.1.
            recall_threshold: Optional; A float controlling the recall score of the classifier. The recall of the trained classifier will be larger than recall_threshold. Default is 0.7.
            n_estimators: Optional; An integer representing the number of trees in random forest model. Default is 500.
            n_neighbors: Optional; An integer representing the number of neighbors in KNN model. Default is 5.

        Returns:
            A dictionary stores the classifier performance on the test set (if test_size is valid).
        """

        if method not in ["RandomForest", "GBDT", "KNN", "NaiveBayes"]:
            msg = "Only support RandomForest, GBDT, KNN, and NaiveBayes method."
            logging.error(msg)
            raise ValueError(msg)

        if valid_size <= 0.0 or valid_size >= 1.0:
            msg = "valid size should be in (0.0, 1.0)"
            logging.error(msg)
            raise ValueError(msg)

        if test_size <= 0.0 or test_size >= 1.0:
            msg = f"invalid test_size={test_size} and reset the test_size as 0."
            test_size = 0.0
            logging.warning(msg)

        n = len(self.features)
        x_train, x_valid, y_train, y_valid = train_test_split(
            self.features, self.labels, test_size=int(n * valid_size)
        )

        if test_size > 0 and test_size < (1 - valid_size):
            x_train, x_test, y_train, y_test = train_test_split(x_train, y_train, test_size=int(n * test_size))
        elif test_size == 0:
            x_train, y_train = self.features, self.labels
            x_test, y_test = None, None
        else:
            msg = "Invalid test_size and re-set test_size as 0."
            logging.info(msg)
            x_train, y_train = self.features, self.labels
            x_test, y_test = None, None
        if method == "NaiveBayes":
            clf = GaussianNB(**kwargs)
        elif method == "GBDT":
            clf = GradientBoostingClassifier(**kwargs)
        elif method == "KNN":
            kwargs["n_neighbors"] = kwargs.get("n_neighbors", 5)
            clf = KNeighborsClassifier(**kwargs)
        else:
            kwargs["n_estimators"] = n_estimators
            kwargs["class_weight"] = kwargs.get("class_weight", "balanced_subsample")
            clf = RandomForestClassifier(**kwargs)
        clf.fit(x_train, y_train)
        pred_valid = clf.predict_proba(x_valid)[:, 1]
        p, r, threshold = precision_recall_curve(y_valid, pred_valid)
        try:
            clf_threshold = threshold[np.where(p == np.max(p[r >= recall_threshold]))][-1]
        except Exception as e:
            msg = f"Fail to get a proper threshold for recall {recall_threshold}, use 0.5 as threshold instead. Exception message is: {e}"
            logging.warning(msg)
            clf_threshold = 0.5
        if x_test is not None:
            pred_test = clf.predict_proba(x_test)[:, 1] > clf_threshold
            precision_test, recall_test, f1_test, _ = precision_recall_fscore_support(
                y_test, pred_test, average="binary"
            )
            accuracy = np.average(pred_test == y_test)
            ans = {
                "accuracy": accuracy,
                "precision": precision_test,
                "recall": recall_test,
                "f1": f1_test,
            }
        else:
            ans = {}
        self.clf = clf
        self._clf_threshold = clf_threshold
        return ans

    def pred(self, source_ts: TimeSeriesData, ts_rescale: bool = True) -> bool:
        """Predict whether a time series is predicable or not.

        Args:
            source_ts: :class:`kats.consts.TimeSeriesData` object representing the new time series data.
            ts_scale: Optional; A boolean to specify whether or not to rescale time series data (i.e., normalizing it with its maximum vlaue) before calculating features. Default is True.

        Returns:
            A boolean representing whether the time series is predictable or not.
        """

        ts = TimeSeriesData(pd.DataFrame(source_ts.to_dataframe().copy()))
        if self.clf is None:
            msg = "No model trained yet, please train the model first."
            logging.error(msg)
            raise ValueError(msg)
        if ts_rescale:
            ts.value /= ts.value.max()
            msg = "Successful scaled! Each value of TS has been divided by the max value of TS."
            logging.info(msg)
        features = TsFeatures().transform(ts)
        # pyre-fixme[16]: `List` has no attribute `values`.
        x = np.array(list(features.values()))
        if np.sum(np.isnan(x)) > 0:
            msg = (
                "Features of ts contain NaN, please consider preprocessing ts. Features are: "
                f"{features}. Fill in NaNs with 0."
            )
            logging.warning(msg)
        ans = True if self.pred_by_feature([x])[0] == 1 else False
        return ans

    def pred_by_feature(self, source_x: Union[np.ndarray, List[np.ndarray], pd.DataFrame]) -> np.ndarray:
        """Predict whether a list of time series are predicable or not given their time series features.
        Args:
            source_x: the time series features of the time series that one wants to predict, can be a np.ndarray, a list of np.ndarray or a pd.DataFrame.

        Returns:
            A np.array storing whether the corresponding time series are predictable or not.
        """

        if self.clf is None:
            msg = "No model trained yet, please train the model first."
            logging.error(msg)
            raise ValueError(msg)
        if isinstance(source_x, List):
            x = np.row_stack(source_x)
        elif isinstance(source_x, np.ndarray):
            x = source_x.copy()
        else:
            msg = f"In valid source_x type: {type(source_x)}."
            logging.error(msg)
            raise ValueError(msg)
        x[np.isnan(x)] = 0.0
        if self.rescale:
            x = (x - self.features_mean) / self.features_std
        pred = (self.clf.predict_proba(x)[:, 1] < self._clf_threshold).astype(int)
        return pred

    def save_model(self, file_path: str) -> None:
        """Save the trained model.

        Args:
            file_name: A string representing the path to save the trained model.

        Returns:
            None.
        """
        if self.clf is None:
            msg = "Please train the model first!"
            logging.error(msg)
            raise ValueError(msg)
        joblib.dump(self.__dict__, file_path)
        logging.info(f"Successfully save the model: {file_path}.")

    # pyre-fixme[2]: Parameter must be annotated.
    def load_model(self, file_path) -> None:
        """Load a pre-trained model.

        Args:
            file_name: A string representing the path to load the pre-trained model.

        Returns:
            None.
        """
        try:
            self.__dict__ = joblib.load(file_path)
            logging.info(f"Successfully load the model: {file_path}.")
        except Exception as e:
            msg = f"Fail to load model with Exception msg: {e}"
            logging.exception(msg)
            raise ValueError(msg)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from unittest import TestCase

from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.models.quadratic_model import QuadraticModel
from kats.models.quadratic_model import QuadraticModelParams


class QuadraticModelTest(TestCase):
    def setUp(self) -> None:
        self.TSData = load_air_passengers()

        DATA_daily = load_data("peyton_manning.csv")
        DATA_daily.columns = ["time", "y"]
        self.TSData_daily = TimeSeriesData(DATA_daily)

        DATA_multi = load_data("multivariate_anomaly_simulated_data.csv")
        self.TSData_multi = TimeSeriesData(DATA_multi)

    def test_fit_forecast(self) -> None:
        params = QuadraticModelParams()
        m = QuadraticModel(self.TSData, params)
        m.fit()
        m.predict(steps=30, freq="MS")
        m.plot()

        m.predict(steps=30, freq="MS", include_history=True)
        m.plot()

        m_daily = QuadraticModel(self.TSData_daily, params)
        m_daily.fit()
        m_daily.predict(steps=30, freq="D")
        m.plot()

        m_daily.predict(steps=30, freq="D", include_history=True)
        m.plot()

    def test_predict_before_fit(self) -> None:
        params = QuadraticModelParams()
        m = QuadraticModel(self.TSData, params)
        with self.assertRaises(ValueError):
            m.predict(steps=30, freq="MS", include_history=True)

    def test_others(self) -> None:
        params = QuadraticModelParams()
        params.validate_params()
        self.assertRaises(
            ValueError,
            QuadraticModel,
            self.TSData_multi,
            params,
        )

        m = QuadraticModel(self.TSData, params)

        # test __str__ method
        self.assertEqual(m.__str__(), "Quadratic")

        # test search space
        self.assertEqual(
            m.get_parameter_search_space(),
            [
                {
                    "name": "alpha",
                    "type": "choice",
                    "value_type": "float",
                    "values": [0.01, 0.05, 0.1, 0.25],
                    "is_ordered": True,
                },
            ],
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from datetime import datetime
from datetime import timedelta
from operator import attrgetter
from typing import cast
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.detectors.detector_consts import AnomalyResponse
from kats.detectors.detector_consts import ChangePointInterval
from kats.detectors.detector_consts import ConfidenceBand
from kats.detectors.detector_consts import PercentageChange
from kats.detectors.detector_consts import SingleSpike


class SingleSpikeTest(TestCase):
    def test_spike(self) -> None:
        spike_time_str = "2020-03-01"
        spike_time = datetime.strptime(spike_time_str, "%Y-%m-%d")
        spike = SingleSpike(time=spike_time, value=1.0, n_sigma=3.0)
        self.assertEqual(spike.time_str, spike_time_str)


class UnivariateChangePointIntervalTest(TestCase):
    # test for univariate time series
    def setUp(self) -> None:
        np.random.seed(100)

        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")
        self.previous_seq = [date_start + timedelta(days=x) for x in range(15)]

        self.current_length = 10

        current_seq = [self.previous_seq[10] + timedelta(days=x) for x in range(self.current_length)]
        self.previous_values = np.random.randn(len(self.previous_seq))
        current_values = np.random.randn(len(current_seq))

        # add a very large value to detect spikes
        current_values[0] = 100.0

        previous = TimeSeriesData(pd.DataFrame({"time": self.previous_seq, "value": self.previous_values}))

        current = TimeSeriesData(pd.DataFrame({"time": current_seq, "value": current_values}))
        self.current_mean = np.mean(current_values)
        self.current_variance = np.var(current_values, ddof=1)

        previous_extend = TimeSeriesData(
            pd.DataFrame({"time": self.previous_seq[9:], "value": self.previous_values[9:]})
        )

        prev_start = self.previous_seq[0]
        prev_end = self.previous_seq[9]

        self.current_start = current_seq[0]
        self.current_start_time_str = datetime.strftime(self.current_start, "%Y-%m-%d")
        self.current_end = current_seq[-1] + timedelta(days=1)
        self.current_end_time_str = datetime.strftime(self.current_end, "%Y-%m-%d")

        self.previous_int = ChangePointInterval(prev_start, prev_end)
        self.previous_int.data = previous

        # construct interval to test whether data is clipped property to start and end dates
        self.previous_int_test = ChangePointInterval(prev_start, prev_end)
        self.previous_int_test.data = previous

        # test extending the data
        # now the data is extended to include the whole sequence
        self.previous_int.end_time = self.previous_seq[-1] + timedelta(days=1)
        self.previous_int.extend_data(previous_extend)

        self.current_int = ChangePointInterval(self.current_start, self.current_end)
        self.current_int.data = current
        self.current_int.previous_interval = self.previous_int

        self.spike_list = self.current_int.spikes

    def test_start_end_date(self) -> None:
        # tests whether data is clipped property to start and end dates
        np.testing.assert_array_equal(self.previous_values[0:9], self.previous_int_test.data)

    def test_interval_seq_length(self) -> None:
        self.assertEqual(len(self.previous_int), len(self.previous_seq))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["start_time", "current_start"],
            ["end_time", "current_end"],
            ["start_time_str", "current_start_time_str"],
            ["end_time_str", "current_end_time_str"],
            ["mean_val", "current_mean"],
            ["variance_val", "current_variance"],
            ["previous_interval", "previous_int"],
        ]
    )
    # check all the properties
    def test_properties(self, attribute: str, initial_object: str) -> None:
        self.assertEqual(attrgetter(attribute)(self.current_int), attrgetter(initial_object)(self))

    def test_length(self) -> None:
        self.assertEqual(len(self.current_int), self.current_length)

    # check spike detection
    def test_spike_start_value(self) -> None:
        self.assertEqual(self.spike_list[0].value, 100.0)

    def test_spike_start_time_str(self) -> None:
        self.assertEqual(
            self.spike_list[0].time_str,
            self.current_start_time_str,
        )


class MultivariateChangePointIntervalTest(TestCase):
    # test for multivariate time series
    def setUp(self) -> None:
        np.random.seed(100)

        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")

        self.previous_seq = [date_start + timedelta(days=x) for x in range(15)]

        self.current_length = 10

        current_seq = [self.previous_seq[10] + timedelta(days=x) for x in range(self.current_length)]

        self.num_seq = 5
        self.previous_values = [np.random.randn(len(self.previous_seq)) for _ in range(self.num_seq)]
        current_values = [np.random.randn(len(current_seq)) for _ in range(self.num_seq)]
        self.current_values_mean = [np.mean(current_values[i]) for i in range(self.num_seq)]
        self.current_values_variance = [np.var(current_values[i]) for i in range(self.num_seq)]

        # add a very large value to detect spikes
        for i in range(self.num_seq):
            current_values[i][0] = 100 * (i + 1)

        previous = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": self.previous_seq},
                    **{f"value_{i}": self.previous_values[i] for i in range(self.num_seq)},
                }
            )
        )

        current = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": current_seq},
                    **{f"value_{i}": current_values[i] for i in range(self.num_seq)},
                }
            )
        )

        previous_extend = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": self.previous_seq[9:]},
                    **{f"value_{i}": self.previous_values[i][9:] for i in range(self.num_seq)},
                }
            )
        )

        prev_start = self.previous_seq[0]
        prev_end = self.previous_seq[9]

        #  `current_start`.
        self.current_start = current_seq[0]
        self.current_start_date_str = datetime.strftime(self.current_start, "%Y-%m-%d")

        self.current_end = current_seq[-1] + timedelta(days=1)
        self.current_end_date_str = datetime.strftime(self.current_end, "%Y-%m-%d")

        self.previous_int = ChangePointInterval(prev_start, prev_end)
        self.previous_int.data = previous

        # now the data is extended to include the whole sequence except the last point
        self.previous_int.end_time = self.previous_seq[-1]  # + timedelta(days=1)
        self.previous_int.extend_data(previous_extend)

        # construct data to test if clipped properly to start and end dates
        self.previous_int1 = ChangePointInterval(prev_start, prev_end)
        self.previous_int1.data = previous

        # let's repeat this except without truncating the final point
        self.previous_int2 = ChangePointInterval(prev_start, prev_end)
        self.previous_int2.data = previous
        self.previous_int2.end_time = self.previous_seq[-1] + timedelta(days=1)
        self.previous_int2.extend_data(previous_extend)

        # let's extend the date range so it's longer than the data
        # this should not change the results
        self.previous_int3 = ChangePointInterval(prev_start, prev_end)
        self.previous_int3.data = previous
        self.previous_int3.end_time = self.previous_seq[-1] + timedelta(days=2)
        self.previous_int3.extend_data(previous_extend)

        # let's construct the current ChangePointInterval
        self.current_int = ChangePointInterval(self.current_start, self.current_end)
        self.current_int.data = current
        self.current_int.previous_interval = self.previous_int
        self.current_int_length = len(self.current_int)

        # spike detection
        self.spike_array = self.current_int.spikes

    def test_start_end_date(self) -> None:
        # tests whether data is clipped properly to start and end dates
        for i in range(self.num_seq):
            self.assertEqual(
                # pyre-fixme[16]: Optional type has no attribute `__getitem__`.
                self.previous_int1.data[:, i].tolist(),
                self.previous_values[i][0:9].tolist(),
            )

    def test_extend_length_except_last_point(self) -> None:
        # test extending the data to include the whole sequence except the last point
        self.assertEqual(len(self.previous_int) + 1, len(self.previous_seq))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["previous_int2"],
    #  ["previous_int3"]])`.
    @parameterized.expand(
        [
            ["previous_int2"],
            ["previous_int3"],
        ]
    )
    def test_extend_length(self, attribute: str) -> None:
        self.assertEqual(len(attrgetter(attribute)(self)), len(self.previous_seq))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["start_time", "current_start"],
            ["end_time", "current_end"],
            ["num_series", "num_seq"],
            ["start_time_str", "current_start_date_str"],
            ["end_time_str", "current_end_date_str"],
            ["previous_interval", "previous_int"],
        ]
    )
    def check_current_int_properties(self, attribute: str, initial_object: str) -> None:
        # check all the properties
        self.assertEqual(attrgetter(attribute)(self.current_int), attrgetter(initial_object)(self))

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["mean_val",
    #  "current_values_mean"], ["variance_val", "current_values_variance"]])`.
    @parameterized.expand(
        [
            ["mean_val", "current_values_mean"],
            ["variance_val", "current_values_variance"],
        ]
    )
    def check_current_int_mean_var(self, attribute: str, initial_object: str) -> None:
        self.assertEqual(
            attrgetter(attribute)(self.current_int).tolist(),
            attrgetter(initial_object)(self),
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["current_int",
    #  "current_length"], ["spike_array", "num_seq"]])`.
    @parameterized.expand([["current_int", "current_length"], ["spike_array", "num_seq"]])
    def check_length(self, attribute: str, initial_object: str) -> None:
        self.assertEqual(len(attrgetter(attribute)(self)), attrgetter(initial_object)(self))

    def check_spike_array_value(self) -> None:
        for i in range(self.num_seq):
            self.assertEqual(self.spike_array[i][0].value, 100 * (i + 1))

    def check_spike_array_time_str(self) -> None:
        for i in range(self.num_seq):
            self.assertEqual(
                self.spike_array[i][0].time_str,
                self.current_start_date_str,
            )


class UnivariatePercentageChangeTest(TestCase):
    # test for univariate time series
    def setUp(self) -> None:
        np.random.seed(100)

        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")
        previous_seq = [date_start + timedelta(days=x) for x in range(30)]

        current_length = 31
        # offset one to make the new interval start one day after the previous one ends
        current_seq = [previous_seq[-1] + timedelta(days=(x + 1)) for x in range(current_length)]
        previous_values = 1.0 + 0.25 * np.random.randn(len(previous_seq))
        current_values = 10.0 + 0.25 * np.random.randn(len(current_seq))

        previous = TimeSeriesData(pd.DataFrame({"time": previous_seq, "value": previous_values}))

        current = TimeSeriesData(pd.DataFrame({"time": current_seq, "value": current_values}))

        previous_int = ChangePointInterval(previous_seq[0], (previous_seq[-1] + timedelta(days=1)))
        previous_int.data = previous

        current_int = ChangePointInterval(current_seq[0], (current_seq[-1] + timedelta(days=1)))
        current_int.data = current
        current_int.previous_interval = previous_int

        self.perc_change_1 = PercentageChange(current=current_int, previous=previous_int)

        previous_mean = np.mean(previous_values)
        current_mean = np.mean(current_values)

        # test the ratios
        self.ratio_val_1 = current_mean / previous_mean

        # test a detector with false stat sig
        second_values = 10.005 + 0.25 * np.random.randn(len(previous_seq))
        second = TimeSeriesData(pd.DataFrame({"time": previous_seq, "value": second_values}))

        second_int = ChangePointInterval(previous_seq[0], previous_seq[-1])
        second_int.data = second

        self.perc_change_2 = PercentageChange(current=current_int, previous=second_int)

        # test the edge case when one of the intervals
        # contains a single data point
        current_int_2 = ChangePointInterval(current_seq[0], current_seq[1])

        current_int_2.data = current

        self.perc_change_3 = PercentageChange(current=current_int_2, previous=previous_int)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["perc_change_1", True],
    #  ["perc_change_2", False]])`.
    @parameterized.expand([["perc_change_1", True], ["perc_change_2", False]])
    def test_stat_sig(self, obj: str, ans: bool) -> None:
        self.assertEqual(attrgetter(obj)(self).stat_sig, ans)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["perc_change_1", True],
    #  ["perc_change_2", False]])`.
    @parameterized.expand([["perc_change_1", True], ["perc_change_2", False]])
    def test_p_value(self, obj: str, ans: bool) -> None:
        self.assertEqual(attrgetter(obj)(self).p_value < 0.05, ans)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["perc_change_1", True],
    #  ["perc_change_2", False], ["perc_change_3", True]])`.
    @parameterized.expand(
        [
            ["perc_change_1", True],
            ["perc_change_2", False],
            ["perc_change_3", True],
        ]
    )
    def test_score(self, obj: str, ans: bool) -> None:
        self.assertEqual(attrgetter(obj)(self).score > 1.96, ans)

    def test_ratio_estimate(self) -> None:
        self.assertEqual(self.perc_change_1.ratio_estimate, self.ratio_val_1)

    def test_approx_ratio_estimate(self) -> None:
        self.assertAlmostEqual(cast(float, self.perc_change_1.ratio_estimate), 10.0, 0)

    def test_direction(self) -> None:
        self.assertEqual(self.perc_change_1.direction, "up")

    def test_perc_change(self) -> None:
        self.assertEqual(self.perc_change_1.perc_change, (self.ratio_val_1 - 1) * 100)

    # TODO delta method tests


class MultivariatePercentageChangeTest(TestCase):
    # test for multivariate time series
    def setUp(self) -> None:
        np.random.seed(100)

        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")
        previous_seq = [date_start + timedelta(days=x) for x in range(30)]

        current_length = 31
        # offset one to make the new interval start one day after the previous one ends
        current_seq = [previous_seq[-1] + timedelta(days=(x + 1)) for x in range(current_length)]

        self.num_seq = 5

        previous_values = np.array([1.0 + 0.0001 * np.random.randn(len(previous_seq)) for _ in range(self.num_seq)])
        current_values = np.array([10.0 + 0.0001 * np.random.randn(len(current_seq)) for _ in range(self.num_seq)])

        previous = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": previous_values[i] for i in range(self.num_seq)},
                }
            )
        )

        current = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": current_seq},
                    **{f"value_{i}": current_values[i] for i in range(self.num_seq)},
                }
            )
        )

        previous_int = ChangePointInterval(previous_seq[0], previous_seq[-1] + timedelta(days=1))
        previous_int.data = previous
        current_int = ChangePointInterval(current_seq[0], current_seq[-1] + timedelta(days=1))
        current_int.data = current
        current_int.previous_interval = previous_int

        self.perc_change_1 = PercentageChange(current=current_int, previous=previous_int)

        previous_mean = np.array([np.mean(previous_values[i]) for i in range(self.num_seq)])
        current_mean = np.array([np.mean(current_values[i]) for i in range(self.num_seq)])

        # test the ratios
        self.ratio_val_1 = current_mean / previous_mean

        # test a detector with false stat sig
        second_values = np.array([10.005 + 0.25 * np.random.randn(len(previous_seq)) for _ in range(self.num_seq)])

        second = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": second_values[i] for i in range(self.num_seq)},
                }
            )
        )

        second_int = ChangePointInterval(previous_seq[0], previous_seq[-1])
        second_int.data = second

        self.perc_change_2 = PercentageChange(current=current_int, previous=second_int)

        # test a detector with a negative spike
        third_values = np.array([1000.0 + 0.0001 * np.random.randn(len(previous_seq)) for _ in range(self.num_seq)])

        third = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": third_values[i] for i in range(self.num_seq)},
                }
            )
        )

        third_int = ChangePointInterval(previous_seq[0], previous_seq[-1])
        third_int.data = third

        self.perc_change_3 = PercentageChange(current=current_int, previous=third_int)

        # test the edge case when one of the intervals
        # contains a single data point
        current_int_single_point = ChangePointInterval(current_seq[0], current_seq[1])

        current_int_single_point.data = current

        self.perc_change_single_point = PercentageChange(current=current_int_single_point, previous=previous_int)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["perc_change_1", True],
    #  ["perc_change_2", False], ["perc_change_3", True], ["perc_change_single_point",
    #  True]])`.
    @parameterized.expand(
        [
            ["perc_change_1", True],
            ["perc_change_2", False],
            ["perc_change_3", True],
            ["perc_change_single_point", True],
        ]
    )
    def test_p_value(self, obj: str, ans: bool) -> None:
        self.assertListEqual(
            # pyre-fixme[16]: `bool` has no attribute `tolist`.
            (attrgetter(obj)(self).p_value < 0.05).tolist(),
            [ans] * self.num_seq,
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["perc_change_1", True],
    #  ["perc_change_2", False]])`.
    @parameterized.expand(
        [
            ["perc_change_1", True],
            ["perc_change_2", False],
        ]
    )
    def test_stat_sig(self, obj: str, ans: bool) -> None:
        self.assertListEqual((attrgetter(obj)(self).stat_sig).tolist(), [ans] * self.num_seq)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["perc_change_1", True],
    #  ["perc_change_2", False], ["perc_change_single_point", True]])`.
    @parameterized.expand(
        [
            ["perc_change_1", True],
            ["perc_change_2", False],
            ["perc_change_single_point", True],
        ]
    )
    def test_score(self, obj: str, ans: bool) -> None:
        self.assertListEqual(
            # pyre-fixme[16]: `bool` has no attribute `tolist`.
            (attrgetter(obj)(self).score > 1.96).tolist(),
            [ans] * self.num_seq,
        )

    def test_score_negative(self) -> None:
        self.assertListEqual(
            # pyre-fixme[16]: `bool` has no attribute `tolist`.
            (self.perc_change_3.score < -1.96).tolist(),
            [True] * self.num_seq,
        )

    def test_approx_ratio_estimate(self) -> None:
        # pyre-fixme[16]: Item `float` of `Union[float, ndarray]` has no attribute
        #  `__iter__`.
        for r in self.perc_change_1.ratio_estimate:
            self.assertAlmostEqual(r, 10.0, 0)

    def test_direction(self) -> None:
        # pyre-fixme[16]: Item `str` of `Union[ndarray, str]` has no attribute `tolist`.
        self.assertEqual(self.perc_change_1.direction.tolist(), ["up"] * self.num_seq)

    def test_perc_change(self) -> None:
        self.assertListEqual(
            # pyre-fixme[16]: `float` has no attribute `tolist`.
            self.perc_change_1.perc_change.tolist(),
            # pyre-fixme[16]: `int` has no attribute `tolist`.
            ((self.ratio_val_1 - 1) * 100).tolist(),
        )


class TestUnivariateAnomalyResponse(TestCase):
    # test anomaly response for univariate time series
    def setUp(self) -> None:
        np.random.seed(100)

        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")
        previous_seq = [date_start + timedelta(days=x) for x in range(30)]
        self.score_ts = TimeSeriesData(
            pd.DataFrame({"time": previous_seq, "value": np.random.randn(len(previous_seq))})
        )
        upper_values = 1.0 + np.random.randn(len(previous_seq))
        upper_ts = TimeSeriesData(pd.DataFrame({"time": previous_seq, "value": upper_values}))

        lower_ts = TimeSeriesData(pd.DataFrame({"time": previous_seq, "value": (upper_values - 0.1)}))

        self.conf_band = ConfidenceBand(upper=upper_ts, lower=lower_ts)

        self.pred_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    "time": previous_seq,
                    "value": (10.0 + 0.25 * np.random.randn(len(previous_seq))),
                }
            )
        )

        self.mag_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    "time": previous_seq,
                    "value": (10.0 + 0.25 * np.random.randn(len(previous_seq))),
                }
            )
        )

        self.stat_sig_ts = TimeSeriesData(pd.DataFrame({"time": previous_seq, "value": np.ones(len(previous_seq))}))

        self.response = AnomalyResponse(
            scores=self.score_ts,
            confidence_band=self.conf_band,
            predicted_ts=self.pred_ts,
            anomaly_magnitude_ts=self.mag_ts,
            stat_sig_ts=self.stat_sig_ts,
        )

        new_date = previous_seq[-1] + timedelta(days=1)
        self.N = len(previous_seq)
        common_val = 1.23

        self.response.update(
            time=new_date,
            score=common_val,
            ci_upper=common_val,
            ci_lower=(common_val - 0.1),
            pred=common_val,
            anom_mag=common_val,
            stat_sig=0,
        )

    def test_response_univariate(self) -> None:
        #  Ensure that num_series is properly populated - this response object is univariate
        self.assertEqual(self.response.num_series, 1)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["scores"],
    #  ["confidence_band.upper"], ["confidence_band.lower"], ["predicted_ts"],
    #  ["anomaly_magnitude_ts"], ["stat_sig_ts"]])`.
    @parameterized.expand(
        [
            ["scores"],
            ["confidence_band.upper"],
            ["confidence_band.lower"],
            ["predicted_ts"],
            ["anomaly_magnitude_ts"],
            ["stat_sig_ts"],
        ]
    )
    def test_update_response_preserves_length(self, attribute: str) -> None:
        # assert that all the lengths of the time series are preserved
        self.assertEqual(len(attrgetter(attribute)(self.response)), self.N)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["scores"],
    #  ["confidence_band.upper"], ["confidence_band.lower"], ["predicted_ts"],
    #  ["anomaly_magnitude_ts"], ["stat_sig_ts"]])`.
    @parameterized.expand(
        [
            ["scores"],
            ["confidence_band.upper"],
            ["confidence_band.lower"],
            ["predicted_ts"],
            ["anomaly_magnitude_ts"],
            ["stat_sig_ts"],
        ]
    )
    def test_get_last_n_length(self, attribute: str) -> None:
        n_val = 10
        response_last_n = self.response.get_last_n(n_val)
        self.assertEqual(len(attrgetter(attribute)(response_last_n)), n_val)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["scores", "score_ts"],
            ["confidence_band.upper", "conf_band.upper"],
            ["confidence_band.lower", "conf_band.lower"],
            ["predicted_ts", "pred_ts"],
            ["anomaly_magnitude_ts", "mag_ts"],
            ["stat_sig_ts", "stat_sig_ts"],
        ]
    )
    def test_update_one_point_forward(self, attribute: str, initial_object: str) -> None:
        self.assertEqual(
            attrgetter(attribute)(self.response).value[0],
            attrgetter(initial_object)(self).value[1],
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["scores", 1.23],  # common_val
            ["confidence_band.upper", 1.23],  # common_val
            ["confidence_band.lower", 1.13],  # common_val-0.1
            ["predicted_ts", 1.23],  # common_val
            ["anomaly_magnitude_ts", 1.23],  # common_val
            ["stat_sig_ts", 0],  # not stat sig
        ]
    )
    def test_last_point(self, attribute: str, new_value: float) -> None:
        # assert that a new point has been added to the end
        self.assertEqual(attrgetter(attribute)(self.response).value.values[-1], new_value)

    def test_get_last_n_values(self) -> None:
        n_val = 10
        response_last_n = self.response.get_last_n(n_val)

        # assert that we return the last N values
        score_list = self.response.scores.value.values.tolist()
        self.assertEqual(response_last_n.scores.value.values.tolist(), score_list[-n_val:])


class TestMultivariateAnomalyResponse(TestCase):
    # test anomaly response for multivariate time series

    def setUp(self) -> None:
        np.random.seed(100)

        date_start_str = "2020-03-01"
        date_start = datetime.strptime(date_start_str, "%Y-%m-%d")
        self.num_seq = 5

        previous_seq = [date_start + timedelta(days=x) for x in range(30)]

        self.score_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": np.random.randn(len(previous_seq)) for i in range(self.num_seq)},
                }
            )
        )

        upper_values = [1.0 + np.random.randn(len(previous_seq)) for _ in range(self.num_seq)]

        upper_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": upper_values[i] for i in range(self.num_seq)},
                }
            )
        )

        lower_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": upper_values[i] - 0.1 for i in range(self.num_seq)},
                }
            )
        )

        self.conf_band = ConfidenceBand(upper=upper_ts, lower=lower_ts)

        self.pred_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": 10.0 + 0.25 * np.random.randn(len(previous_seq)) for i in range(self.num_seq)},
                }
            )
        )

        self.mag_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": 10.0 + 0.25 * np.random.randn(len(previous_seq)) for i in range(self.num_seq)},
                }
            )
        )

        self.stat_sig_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    **{"time": previous_seq},
                    **{f"value_{i}": np.ones(len(previous_seq)) for i in range(self.num_seq)},
                }
            )
        )

        self.response = AnomalyResponse(
            scores=self.score_ts,
            confidence_band=self.conf_band,
            predicted_ts=self.pred_ts,
            anomaly_magnitude_ts=self.mag_ts,
            stat_sig_ts=self.stat_sig_ts,
        )

        # test update
        new_date = previous_seq[-1] + timedelta(days=1)
        common_val = 1.23 * np.ones(self.num_seq)

        self.response.update(
            time=new_date,
            score=common_val,
            ci_upper=common_val,
            ci_lower=common_val - 0.1,
            pred=common_val,
            anom_mag=common_val,
            stat_sig=np.zeros(self.num_seq),
        )

        self.N = len(previous_seq)

    def test_response_num_series(self) -> None:
        # Ensure that num_series is properly populated
        self.assertEqual(self.response.num_series, self.num_seq)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["scores"],
    #  ["confidence_band.upper"], ["confidence_band.lower"], ["predicted_ts"],
    #  ["anomaly_magnitude_ts"], ["stat_sig_ts"]])`.
    @parameterized.expand(
        [
            ["scores"],
            ["confidence_band.upper"],
            ["confidence_band.lower"],
            ["predicted_ts"],
            ["anomaly_magnitude_ts"],
            ["stat_sig_ts"],
        ]
    )
    def test_update_response_preserves_length(self, attribute: str) -> None:
        # assert that all the lengths of the time series are preserved
        self.assertEqual(len(attrgetter(attribute)(self.response)), self.N)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["scores", "score_ts"],
            ["confidence_band.upper", "conf_band.upper"],
            ["confidence_band.lower", "conf_band.lower"],
            ["predicted_ts", "pred_ts"],
            ["anomaly_magnitude_ts", "mag_ts"],
            ["stat_sig_ts", "stat_sig_ts"],
        ]
    )
    def test_update_one_point_forward(self, attribute: str, initial_object: str) -> None:
        self.assertEqual(
            attrgetter(attribute)(self.response).value.iloc[0].tolist(),
            attrgetter(initial_object)(self).value.iloc[1].tolist(),
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            ["scores", 1.23],  # common_val
            ["confidence_band.upper", 1.23],  # common_val
            ["confidence_band.lower", 1.13],  # common_val-0.1
            ["predicted_ts", 1.23],  # common_val
            ["anomaly_magnitude_ts", 1.23],  # common_val
            ["stat_sig_ts", 0],  # not stat sig
        ]
    )
    def test_last_point(self, attribute: str, new_value: float) -> None:
        # assert that a new point has been added to the end
        self.assertEqual(
            attrgetter(attribute)(self.response).value.iloc[-1].tolist(),
            (np.ones(self.num_seq) * new_value).tolist(),
        )

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator
    #  `parameterized.parameterized.parameterized.expand([["scores"],
    #  ["confidence_band.upper"], ["confidence_band.lower"], ["predicted_ts"],
    #  ["anomaly_magnitude_ts"], ["stat_sig_ts"]])`.
    @parameterized.expand(
        [
            ["scores"],
            ["confidence_band.upper"],
            ["confidence_band.lower"],
            ["predicted_ts"],
            ["anomaly_magnitude_ts"],
            ["stat_sig_ts"],
        ]
    )
    def test_get_last_n_length(self, attribute: str) -> None:
        n_val = 10
        response_last_n = self.response.get_last_n(n_val)
        self.assertEqual(len(attrgetter(attribute)(response_last_n)), n_val)

    def test_get_last_n_values(self) -> None:
        n_val = 10
        response_last_n = self.response.get_last_n(n_val)

        # assert that we return the last N values
        score_list = self.response.scores.value.values.tolist()
        self.assertEqual(response_last_n.scores.value.values.tolist(), score_list[-n_val:])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any
from typing import Dict
from unittest import TestCase

import numpy as np
import pandas as pd

from kats.compat.pandas import assert_frame_equal
from kats.utils import feature_engineering as fe


class FeatureEngineeringTest(TestCase):
    def assertDictAlmostEqual(self, expected: Dict[str, Any], features: Dict[str, Any], places: int = 4) -> None:
        """Compares that two dictionaries are floating-point almost equal.

        Note: the dictionaries may or may not contain floating-point values.

        Args:
          expected: the expected dictionary of values.
          actual: the actual dictionary of values.
          places: the number of decimal places for floating point comparisons.
        """

        self.assertEqual(expected.keys(), features.keys())
        for k, v in expected.items():
            if isinstance(v, float):
                if np.isnan(v):
                    self.assertTrue(np.isnan(features[k]), msg=f"{k} differ")
                else:
                    self.assertAlmostEqual(v, features[k], places=4, msg=f"{k} differ")
            else:
                self.assertEqual(v, features[k], msg=f"{k} differ")

    def test_datetime_features(self) -> None:
        dates = pd.date_range("2021-01-22", "2021-01-31", tz="US/Pacific").tolist()
        dates += pd.date_range("2022-01-30", "2022-02-04", tz="US/Pacific").tolist()
        values = [10, 2, 11, 5, 13, 18, 4, 14, 17, 9] + [np.nan] * 6
        self.assertEqual(len(values), len(dates))
        expected = pd.DataFrame(
            {
                "val": values,
                "year": [2021] * 10 + [2022] * 6,
                "month": [1] * 12 + [2] * 4,
                "day": list(range(22, 32)) + [30, 31, 1, 2, 3, 4],
                "dayofweek": [4, 5, 6] + list(range(7)) + [6, 0, 1, 2, 3, 4],
                "dayofyear": list(range(22, 32)) + list(range(30, 36)),
                "quarter": [1] * 16,
                "season": [0] * 12 + [1] * 4,
                "weekofyear": [3] * 3 + [4] * 8 + [5] * 5,
                "weekofmonth": [4] * 3 + [5] * 8 + [6, 1, 1, 1, 1],
                "is_weekend": [False, True, True] + [False] * 5 + [True] * 3 + [False] * 5,
                "is_leap_year": [False] * 16,
                "is_leap_day": [False] * 16,
                "is_month_end": [False] * 9 + [True, False, True] + [False] * 4,
                "is_quarter_end": [False] * 16,
                "hour": [0] * 16,
                "minute": [0] * 16,
                "second": [0] * 16,
                "milliseconds": [0.0] * 16,
                "quarterhour": [1] * 16,
                "hourofweek": [96, 120, 144] + list(range(0, 168, 24)) + [144, 0, 24, 48, 72, 96],
                "daytime": [0] * 16,
                "dst": [0.0] * 16,
                "utcoffset": [-28800.0] * 16,
            },
            index=dates,
        )
        result = fe.datetime_features(pd.Series(values, index=dates, name="val"))
        assert_frame_equal(expected, result)

    def test_date_features(self) -> None:
        dates = pd.date_range("2021-01-22", "2021-01-31", tz="US/Pacific").tolist()
        dates += pd.date_range("2022-01-30", "2022-02-04", tz="US/Pacific").tolist()
        values = [10, 2, 11, 5, 13, 18, 4, 14, 17, 9] + [np.nan] * 6
        self.assertEqual(len(values), len(dates))
        expected = pd.DataFrame(
            {
                "val": values,
                "year": [2021] * 10 + [2022] * 6,
                "month": [1] * 12 + [2] * 4,
                "day": list(range(22, 32)) + [30, 31, 1, 2, 3, 4],
                "dayofweek": [4, 5, 6] + list(range(7)) + [6, 0, 1, 2, 3, 4],
                "dayofyear": list(range(22, 32)) + list(range(30, 36)),
                "quarter": [1] * 16,
                "season": [0] * 12 + [1] * 4,
                "weekofyear": [3] * 3 + [4] * 8 + [5] * 5,
                "weekofmonth": [4] * 3 + [5] * 8 + [6, 1, 1, 1, 1],
                "is_weekend": [False, True, True] + [False] * 5 + [True] * 3 + [False] * 5,
                "is_leap_year": [False] * 16,
                "is_leap_day": [False] * 16,
                "is_month_end": [False] * 9 + [True, False, True] + [False] * 4,
                "is_quarter_end": [False] * 16,
            },
            index=dates,
        )
        result = fe.date_features(pd.Series(values, index=dates, name="val"))
        assert_frame_equal(expected, result)

    def test_time_features(self) -> None:
        dates = pd.date_range("2021-01-22", "2021-01-31", tz="US/Pacific").tolist()
        dates += pd.date_range("2022-01-30", "2022-02-04", tz="US/Pacific").tolist()
        values = [10, 2, 11, 5, 13, 18, 4, 14, 17, 9] + [np.nan] * 6
        self.assertEqual(len(values), len(dates))
        expected = pd.DataFrame(
            {
                "val": values,
                "hour": [0] * 16,
                "minute": [0] * 16,
                "second": [0] * 16,
                "milliseconds": [0.0] * 16,
                "quarterhour": [1] * 16,
                "hourofweek": [96, 120, 144] + list(range(0, 168, 24)) + [144, 0, 24, 48, 72, 96],
                "daytime": [0] * 16,
                "dst": [0.0] * 16,
                "utcoffset": [-28800.0] * 16,
            },
            index=dates,
        )
        result = fe.time_features(pd.Series(values, index=dates, name="val"))
        assert_frame_equal(expected, result)

    def test_timestamp_time_features(self) -> None:
        t = pd.Timestamp("2021-01-01T02:03:04.5678", tz="US/Pacific")
        expected = {
            "year": 2021,
            "month": 1,
            "day": 1,
            "dayofweek": 4,
            "dayofyear": 1,
            "quarter": 1,
            "season": 0,
            "weekofyear": 53,
            "weekofmonth": 1,
            "is_weekend": False,
            "is_leap_year": False,
            "is_leap_day": False,
            "is_month_end": False,
            "is_quarter_end": False,
            "hour": 2,
            "minute": 3,
            "second": 4,
            "milliseconds": 567.8,
            "quarterhour": 1,
            "hourofweek": 98,
            "daytime": 0,
            "dst": 0.0,
            "utcoffset": -28800.0,
        }
        result = fe.timestamp_datetime_features(t)
        self.assertDictAlmostEqual(expected, result)

    def test_timestamp_time_features_notz(self) -> None:
        t = pd.Timestamp("2021-01-01T02:03:04.5678")
        expected = {
            "year": 2021,
            "month": 1,
            "day": 1,
            "dayofweek": 4,
            "dayofyear": 1,
            "quarter": 1,
            "season": 0,
            "weekofyear": 53,
            "weekofmonth": 1,
            "is_weekend": False,
            "is_leap_year": False,
            "is_leap_day": False,
            "is_month_end": False,
            "is_quarter_end": False,
            "hour": 2,
            "minute": 3,
            "second": 4,
            "milliseconds": 567.8,
            "quarterhour": 1,
            "hourofweek": 98,
            "daytime": 0,
            "dst": np.nan,
            "utcoffset": np.nan,
        }
        result = fe.timestamp_datetime_features(t)
        self.assertDictAlmostEqual(expected, result)

    def test_circle_encode(self) -> None:
        s32 = np.sqrt(3) / 2
        expected = pd.DataFrame(
            {
                "x_cos": [0.5, s32, 1.0, s32],
                "x_sin": [-s32, -0.5, 0.0, 0.5],
            }
        )
        result = fe.circle_encode(pd.DataFrame({"x": [10, 11, 12, 1]}), {"x": 12})
        assert_frame_equal(expected, result)

    def test_circle_encode_modulo(self) -> None:
        s32 = np.sqrt(3) / 2
        expected = pd.DataFrame(
            {
                "x_cos": [0.5, s32, 1.0, s32],
                "x_sin": [-s32, -0.5, 0.0, 0.5],
            }
        )
        result = fe.circle_encode(pd.DataFrame({"x": [10, 11, 12, 13]}), {"x": 12}, modulo=True)
        assert_frame_equal(expected, result)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This module contains 1) the BaseTHModel class for storing information of base models;
and 2) the GetAggregateTS class for aggregating base time series to higher levels.
"""

from typing import Dict
from typing import List
from typing import Optional

from kats.consts import TimeSeriesData
from kats.models import arima
from kats.models import holtwinters
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import sarima
from kats.models import theta

BASE_MODELS = {
    "arima": arima.ARIMAModel,
    "holtwinters": holtwinters.HoltWintersModel,
    "sarima": sarima.SARIMAModel,
    "linear": linear_model.LinearModel,
    "quadratic": quadratic_model.QuadraticModel,
    "theta": theta.ThetaModel,
}
import logging

import numpy as np
import pandas as pd


class BaseTHModel:
    """Base class for temporal hierarhical models.

    The object stores the information of base model. We allow users to pass model info (i.e., model_name and model_params),
    or to pass residuals and forecasts of a trained model directly.

    Attributes:
        level: An integer representing the level of the base model, should be a positive integer.
        model_name: Optional; A string representing the name of forecast model; Default is None.
        model_params: Optional; A parameter object storing the parameters of forecasting model; Default is None.
        residuals: Optional; A np.array of residuals of forecasting model, which is necessary if both model_name and model_params are None. Default is None.
        fcsts: Optional; A np.array of forecasts generated by the forecasting model, which is necessary if both model_name and model_params are None. Default is None.
    """

    def __init__(
        self,
        level: int,
        model_name: Optional[str] = None,
        model_params: Optional[object] = None,
        residuals: Optional[np.ndarray] = None,
        fcsts: Optional[np.ndarray] = None,
    ) -> None:

        if not isinstance(level, int) or level < 1:
            msg = f"Level should be a positive integer but receive {level}."
            logging.error(msg)
            raise ValueError(msg)

        if (residuals is None or residuals.size == 0) or (fcsts is None or fcsts.size == 0):
            # when residuals or fcsts are missing
            if (not model_name) or (not model_params):
                msg = "model_name and model_params are needed when residuals or fcsts is missing."
                logging.error(msg)
                raise ValueError(msg)
            if model_name not in BASE_MODELS:
                msg = f"model_name {model_name} is not supported!"
                logging.error(msg)
                raise ValueError(msg)

        self.level = level
        self.model_name = model_name
        self.model_params = model_params
        self.residuals = residuals
        self.fcsts = fcsts

    # pyre-fixme[3]: Return type must be annotated.
    def __str__(self):
        return "BaseTHModel"


class GetAggregateTS:
    """Class for aggregating time series to different levels.

    This class provides aggregate.

    Attributes:
        data: A TimeSeriesData object representing the time series to be aggregated.
    """

    def __init__(self, data: TimeSeriesData) -> None:

        if not data.is_univariate():
            msg = f"Only support univariate time series, but get {type(data.value)}."
            logging.error(msg)
            raise ValueError(msg)

        self.data = TimeSeriesData(data.to_dataframe().copy())

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _aggregate_single(self, ts, k):
        if k == 1:
            return ts
        if k > len(ts):
            msg = f"Level {k} should be less than the length of training time series (len(TS)={len(ts)})!"
            logging.error(msg)
            raise ValueError(msg)
        n = len(ts)
        m = (n // k) * k

        value = pd.Series(ts.value.values[-m:].reshape(-1, k).sum(axis=1), copy=False)
        time = pd.Series(ts.time.values[(n - m + k - 1) : n : k], copy=False)
        return TimeSeriesData(time=time, value=value)

    def aggregate(self, levels: List[int]) -> Dict[int, TimeSeriesData]:
        """Function for aggregating time series.

        Args:
            levels: A list of integers representing the levels which the time series to be aggregated for.

        Returns:
            A dictionary of aggregated time series for each level.
        """

        if not isinstance(levels, list):
            msg = f"Parameter 'levels' should be a list but receive {type(levels)}."
            logging.error(msg)
            raise ValueError(msg)
        for k in levels:
            if not isinstance(k, int) or k < 1:
                msg = f"Level should be a positive int, but receive {k}."
                logging.error(msg)
                raise ValueError(msg)
        return {k: self._aggregate_single(self.data, k) for k in levels}

    # pyre-fixme[3]: Return type must be annotated.
    def __str__(self):
        return "GetAggregateTS"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""NowcastingPlus is a basic model for short-term forecasting.

This modules contains class NowcastingParams, which is the class parameter
and class NowcastingPlusModel, which is the model.

  Typical usage example:

  nr = NowcastingPlusModel(data = data, params = NowcastingParams(step = 10))
  nr.feature_extraction()
  nr.label_extraction()
  nr.fit()
  output = nr.predict()
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
from typing import Any
from typing import List

import numpy as np
import pandas as pd
from sklearn import linear_model
from sklearn import preprocessing
from sklearn.linear_model import LinearRegression

import kats.models.model as m
from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.nowcasting.feature_extraction import LAG
from kats.models.nowcasting.feature_extraction import MA
from kats.models.nowcasting.feature_extraction import MOM
from kats.models.nowcasting.feature_extraction import ROC
from kats.models.nowcasting.model_io import deserialize_from_zippy
from kats.models.nowcasting.model_io import serialize_for_zippy


# pyre-fixme[3]: Return type must be annotated.
# pyre-fixme[2]: Parameter must be annotated.
def poly(df, n):
    """
    Takes the column x from the dataframe df and takes
    the value from x to the power n
    """
    poly = pd.Series(df.x**n, name="poly_" + str(n))
    df = df.join(poly)
    return df


class NowcastingParams(Params):
    """The class for Nowcasting Parameters.

    Takes parameters for class NowcastingModel.

    Attributes:
        step: An integer indicating how many steps ahead we are forecasting. Default is 1.
    """

    # pyre-fixme[2]: Parameter must be annotated.
    def __init__(self, step: int = 1, **kwargs) -> None:
        super().__init__()
        self.step = step
        logging.debug(f"Initialized QuadraticModel with parameters: step:{step}")

    # pyre-fixme[3]: Return type must be annotated.
    def validate_params(self):
        """Raises: NotImplementedError("Subclasses should implement this!")."""

        logging.warning("Method validate_params() is not implemented.")
        raise NotImplementedError("Subclasses should implement this!")


# pyre-fixme[24]: Generic type `m.Model` expects 1 type parameter.
class NowcastingPlusModel(m.Model):
    """The class for NowcastingPlus Model.

    This class performs data processing and short term prediction, for time series
    based on machine learning methodology.

    Attributes:
        TimeSeriesData: Time Series Data Source.
        NowcastingParams: parameters for Nowcasting.
    """

    def __init__(
        self,
        data: TimeSeriesData,
        params: NowcastingParams,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        model: Any = None,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        poly_model: Any = None,
        feature_names: List[str] = [],
        poly_feature_names: List[str] = [],
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        scaler: Any = None,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        label_scaler: Any = None,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        y_train_season_obj: Any = None,
    ) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: Optional type has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)
        # pyre-fixme[4]: Attribute must be annotated.
        self.df = data.to_dataframe()
        # pyre-fixme[4]: Attribute must be annotated.
        self.step = params.step
        self.model = model
        self.feature_names = feature_names
        self.poly_model = poly_model
        # pyre-fixme[4]: Attribute must be annotated.
        self.df_poly = data.to_dataframe()
        self.poly_feature_names = poly_feature_names
        # pyre-fixme[4]: Attribute must be annotated.
        self.df_nowcasting = data.to_dataframe()
        self.scaler = scaler
        self.label_scaler = label_scaler
        self.y_train_season_obj = y_train_season_obj

    def feature_extraction(self) -> None:
        """
        Extracts features for time series data.
        """
        # Add the hour, minute, and x column to the data
        self.df_poly["hour"] = self.df_poly["time"].apply(lambda y: y.hour)
        self.df_poly["minute"] = self.df_poly["time"].apply(lambda y: y.minute)
        self.df_poly["x"] = self.df_poly["hour"] * 60 + self.df_poly["minute"]

        # Empty list to hold the feature names
        poly_feature_names = []

        # Add the poly columns to the df_poly
        for degree in [0, 1, 2, 3, 4, 5]:
            self.df_poly = poly(self.df_poly, degree)
            poly_feature_names.append("poly_" + str(degree))

        # filterout + - inf, nan
        self.df_poly = self.df_poly[~self.df_poly.isin([np.nan, np.inf, -np.inf]).any(1)]

        # Save the poly feature name
        self.poly_feature_names = poly_feature_names
        feature_names = []

        #########################################################################################
        train_index_poly = self.df_poly[~self.df_poly.isin([np.nan, np.inf, -np.inf]).any(1)].index
        X_train_poly, y_train_poly = (
            self.df_poly[self.poly_feature_names].loc[train_index_poly],
            self.df_poly["y"].loc[train_index_poly],
        )

        # Build the Polynomial Regression Model
        lin_reg = LinearRegression()
        lin_reg.fit(X_train_poly, y_train_poly)
        self.poly_model = lin_reg
        y_train_season = lin_reg.predict(X_train_poly)
        self.y_train_season_obj = y_train_season
        #########################################################################################

        for n in [10, 15, 20, 25, 30]:
            self.df = MOM(self.df, n)
            feature_names.append("MOM_" + str(n))
        for n in [10, 15, 20, 25, 30]:
            self.df = ROC(self.df, n)
            feature_names.append("ROC_" + str(n))
        for n in [1, 2, 3, 4, 5]:
            self.df = LAG(self.df, n)
            feature_names.append("LAG_" + str(n))
        for n in [10, 20, 30]:
            self.df = MA(self.df, n)
            feature_names.append("MA_" + str(n))

        self.df = self.df[~self.df.isin([np.nan, np.inf, -np.inf]).any(1)]  # filterout + - inf, nan
        self.feature_names = feature_names

    def label_extraction(self) -> None:
        """Extracts labels from time series data."""
        self.df["label"] = self.df["y"]

    ###################### module 1: for offline training ######################

    def fit(self) -> None:
        """Fits model."""

        logging.debug("Call fit() with parameters: " "step:{step}".format(step=self.step))

        n = 1
        train_index = self.df[~self.df.isin([np.nan, np.inf, -np.inf]).any(1)].index

        X_train = self.df[self.feature_names].loc[train_index]
        std_scaler = preprocessing.StandardScaler()
        X_train = std_scaler.fit_transform(X_train)
        self.scaler = std_scaler

        n = self.step
        y_train = (self.df["label"].loc[train_index] - self.y_train_season_obj[train_index]).diff(-n)[:-n]
        X_train = X_train[:-n]

        reg = linear_model.LassoCV()
        reg.fit(X_train, y_train)
        self.model = reg

    def save_model(self) -> bytes:
        """Saves sklearn model as bytes."""

        return serialize_for_zippy(self.model)

    ###################### module 2: for online prediction ######################
    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def predict(self, **kwargs):
        """Predicts the time series in the future.

        Nowcasting forecasts at the time unit of step ahead.
        This is in order to keep precision and different from usual algorithms.
        Returns:
            A float variable, the forecast at future step.
        """

        logging.debug("Call predict() with parameters. " "Forecast 1 step only, kwargs:{kwargs}".format(kwargs=kwargs))

        X_test = self.df[-self.step :][self.feature_names]
        X_test = self.scaler.transform(X_test)
        y_predict = self.model.predict(X_test)
        poly_now = self.y_train_season_obj[-1]
        first_occ = np.where(self.y_train_season_obj == poly_now)
        polynext = self.y_train_season_obj[first_occ[0][0] + self.step]
        now = self.df["y"][-self.step :]
        return (now - poly_now) - y_predict + polynext

    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def predict_polyfit(self, model=None, df=None, **kwargs):
        poly_now = self.y_train_season_obj[-1]
        first_occ = np.where(self.y_train_season_obj == poly_now)
        polynext = self.y_train_season_obj[first_occ[0][0] + self.step]
        return polynext

    def load_model(self, model_as_bytes: bytes) -> None:
        """Loads model_as_str and decodes into the class NowcastingModel.

        Args:
            model_as_bytes: a binary variable, indicating whether to read as bytes.
        """

        self.model = deserialize_from_zippy(model_as_bytes)

    # pyre-fixme[14]: `plot` overrides method defined in `Model` inconsistently.
    # pyre-fixme[3]: Return type must be annotated.
    def plot(self):
        """Raises: NotImplementedError("Subclasses should implement this!")。"""

        raise NotImplementedError("Subclasses should implement this!")

    # pyre-fixme[3]: Return type must be annotated.
    def __str__(self):
        """Returns the name as Nowcasting,"""

        return "Nowcasting"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from . import backtesters
from . import cupik
from . import decomposition
from . import emp_confidence_int
from . import parameter_tuning_utils
from . import simulator



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pickle
from typing import Any


class SimplePickleSerializer:
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def _jdefault(self, o):
        if isinstance(o, set):
            return list(o)
        if isinstance(o, bool):
            return str(o).lower()
        if isinstance(o, int):
            return str(o)
        if isinstance(o, float):
            return str(o)
        return o.__dict__

    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def serialize(self, obj: Any) -> bytes:
        """Performs model saving.

        Args:
            obj is an object to be saved. Usually it is an sklearn model.

        Returns:
            A bytes object which is the compressed model.
        """

        if obj is None:

            return b""
        return pickle.dumps(obj)

    # pyre-fixme[3]: Return annotation cannot be `Any`.
    def deserialize(self, serialized_data: bytes) -> Any:
        """Performs model decoding.

        Args:
            serialized_data is a bytes object to be decoded.

        Returns:
            A decompressed model. Usually a sklearn model.
        """

        if serialized_data is None:
            return None
        decoded = serialized_data  # .decode("utf-8")
        if not decoded:
            return None
        return pickle.loads(decoded)


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def serialize_for_zippy(input: Any) -> bytes:
    """Performs model compression.

    Args:
        Input is an sklearn model.

    Returns:
        A compressed version of the model.
    """

    serializer = SimplePickleSerializer()
    return serializer.serialize(input)


# pyre-fixme[2]: Parameter must be annotated.
def deserialize_from_zippy(input: bytes, use_case_id=None) -> None:
    """Performs model serialization for Zippydb.

    Args:
        Input is an encoded sklearn model.

    Returns:
        A compressed version of the model.
    """

    serializer = SimplePickleSerializer()
    return serializer.deserialize(input)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Any
from unittest.mock import patch

import numpy as np

from kats.compat import compat
from kats.compat import sklearn


class TestSklearn(unittest.TestCase):
    def setUp(self) -> None:
        y_true = np.array([1.0, 2.0])
        y_pred = np.array([1.0, 1.0])
        sample_weight = np.array([1.0, 1.0])
        self.orig_args = [y_true, y_pred, sample_weight]
        self.args = [y_true, y_pred]
        self.kwargs = {
            "sample_weight": sample_weight,
            "multioutput": "uniform_average",
            "squared": True,
        }

    def test_version(self) -> None:
        self.assertTrue(sklearn.version == compat.Version("sklearn"))

    @patch("kats.compat.sklearn.mse")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_mean_squared_error(self, sklearn_mse: Any) -> None:
        sklearn_mse.return_value = 42.0
        result = sklearn.mean_squared_error(*self.orig_args)
        self.assertEqual(42.0, result)
        sklearn_mse.assert_called_once_with(*self.args, **self.kwargs)

    @patch("kats.compat.sklearn.mse")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_root_mean_squared_error(self, sklearn_mse: Any) -> None:
        sklearn_mse.return_value = 42.0
        result = sklearn.mean_squared_error(*self.orig_args, squared=False)
        # sqrt part of the mocked call
        self.assertEqual(42.0, result)
        kwargs = dict(self.kwargs)
        kwargs["squared"] = False
        sklearn_mse.assert_called_once_with(*self.args, **kwargs)

    @patch("kats.compat.sklearn.version", compat.Version("0.24"))
    @patch("kats.compat.sklearn.msle")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_mean_squared_log_error(self, sklearn_msle: Any) -> None:
        sklearn_msle.return_value = 42.0
        result = sklearn.mean_squared_log_error(*self.orig_args)
        self.assertEqual(42.0, result)
        kwargs = dict(self.kwargs)
        del kwargs["squared"]
        sklearn_msle.assert_called_once_with(*self.args, **kwargs)

    @patch("kats.compat.sklearn.version", compat.Version("0.24"))
    @patch("kats.compat.sklearn.msle")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_root_mean_squared_log_error_24(self, sklearn_msle: Any) -> None:
        sklearn_msle.return_value = 36.0
        result = sklearn.mean_squared_log_error(*self.orig_args, squared=False)
        # sqrt applied after-the-fact
        self.assertEqual(6.0, result)
        kwargs = dict(self.kwargs)
        del kwargs["squared"]
        sklearn_msle.assert_called_once_with(*self.args, **kwargs)

    @patch("kats.compat.sklearn.version", compat.Version("1.0"))
    @patch("kats.compat.sklearn.msle")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_root_mean_squared_log_error(self, sklearn_msle: Any) -> None:
        sklearn_msle.return_value = 36.0
        result = sklearn.mean_squared_log_error(*self.orig_args, squared=False)
        # sqrt part of the mocked call
        self.assertEqual(36.0, result)
        kwargs = dict(self.kwargs)
        kwargs["squared"] = False
        sklearn_msle.assert_called_once_with(*self.args, **kwargs)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Detectors based on predictors, basically work as follows:
calculate the residual (i.e., difference between predicted and current value),
and translate it into a false-alarm probability by how large it is.
This is often done by assuming that residuals are distributed normally.
In practice, the residuals are often non-normal (sometimes even being
asymmetric). This module “learns” the distribution of the residual (using kernel
density estimation), and outputs a false-alarm probability based on it.
"""
from __future__ import annotations

from typing import Optional

import numpy as np
import pandas as pd
from sklearn.model_selection import RandomizedSearchCV
from sklearn.neighbors import KernelDensity

from kats.consts import TimeSeriesData


class KDEResidualTranslator:
    _kde: Optional[KernelDensity] = None

    def __init__(self, ignore_below_frac: float = 0, ignore_above_frac: float = 1) -> None:
        """
        Translates residuals (difference between outcome and prediction)
        to false-alarm probability using kernel density estimation
        on the residuals.

        Args:
            ignore_below_frac: Lower quantile to ignore during training
                (makes the translator more robust to outliers); default 0.
            ignore_above_frac: Upper quantile to ignore during training
                (makes the translator more robust to outliers); default 1.

        Examples:

        .. code-block:: py

            # ts_data is some data
            y = ts_data
            # We create a prediction by a rolling window of length 7
            yhat = pd.DataFrame({
                "value": self._y.value.rolling(7).mean().shift(1),
                "time": self._y.time})
            yhat = TimeSeriesData(yhat)

            trn = KDEResidualTranslator()

            # We can transform outcomes to probabilities using y and y_hat
            trn = trn.fit(y=y, yhat=yhat)
            proba = trn.predict(y)

            # We can transform outcomes to probabilities using residuals
            residual = self._y - self._yhat
            trn = trn.fit(residual=residual)
            proba = trn.predict(residual=residual)
        """
        if ignore_below_frac < 0 or ignore_above_frac > 1:
            raise ValueError("Illegal ignore fractions")
        if ignore_below_frac > ignore_above_frac:
            raise ValueError("Illegal ignore fractions")
        self._ignore_below_frac = ignore_below_frac
        self._ignore_above_frac = ignore_above_frac

    def fit(
        self,
        y: Optional[TimeSeriesData] = None,
        yhat: Optional[TimeSeriesData] = None,
        yhat_lower: Optional[TimeSeriesData] = None,
        yhat_upper: Optional[TimeSeriesData] = None,
        residual: Optional[TimeSeriesData] = None,
    ) -> KDEResidualTranslator:
        """
        Fits a dataframe to a model of the residuals.

        Args:
            df: A pandas DataFrame containg the following columns:

                1. Either
                    a. `residual`, or
                    b. `y` and `yhat` with optionally both `yhat_lower` and
                       `yhat_upper`
                2. At most one of `ds` and `ts`
        """
        residual = self._get_residual(y, yhat, yhat_lower, yhat_upper, residual)

        value = residual.value
        mask = value > value.quantile(self._ignore_below_frac)
        # pyre-fixme[58]: `&` is not supported for operand types
        #  `Union[pd.core.frame.DataFrame, pd.core.series.Series]` and
        #  `Union[pd.core.frame.DataFrame, pd.core.series.Series]`.
        mask &= value < value.quantile(self._ignore_above_frac)
        value = value[mask]

        kde = KernelDensity(bandwidth=10.0, kernel="gaussian")
        extent = value.quantile(0.95) - value.quantile(0.05)
        params = {
            "kernel": ["gaussian"],
            "bandwidth": np.linspace(extent / 1000, extent / 10, 1000),
        }
        search = RandomizedSearchCV(kde, params, random_state=0, scoring=lambda k, x: k.score_samples(x).sum())
        best_params = search.fit(value.to_frame()).best_params_
        kde = KernelDensity(**best_params)
        kde.fit(value.to_frame())
        self._kde = kde
        return self

    @property
    def kde_(self) -> Optional[KernelDensity]:
        """
        Returns:
            KernelDensity object fitted to the residuals.
        """
        return self._kde

    def predict_proba(
        self,
        y: Optional[TimeSeriesData] = None,
        yhat: Optional[TimeSeriesData] = None,
        yhat_lower: Optional[TimeSeriesData] = None,
        yhat_upper: Optional[TimeSeriesData] = None,
        residual: Optional[TimeSeriesData] = None,
    ) -> TimeSeriesData:
        """
        Predicts the probability of a residual

        Args:
            df: A pandas DataFrame containg the following columns:

                1. Either
                    a. `residual`, or
                    b. `y` and `yhat` with optionally both `yhat_lower` and
                        `yhat_upper`
                2. At most one of `ds` and `ts`

        Returns:
            A series where there is a probability corresponding to
                each instance (row) in the input.
        """
        proba = self.predict_log_proba(y, yhat, yhat_lower, yhat_upper, residual)
        proba.value = np.exp(proba.value)
        return proba

    def predict_log_proba(
        self,
        y: Optional[TimeSeriesData] = None,
        yhat: Optional[TimeSeriesData] = None,
        yhat_lower: Optional[TimeSeriesData] = None,
        yhat_upper: Optional[TimeSeriesData] = None,
        residual: Optional[TimeSeriesData] = None,
    ) -> TimeSeriesData:
        """
        Predicts the natural-log probability of a residual

        Args:
            df: A pandas DataFrame containg the following columns:

                1. Either
                    a. `residual`, or
                    b. `y` and `yhat` with optionally both `yhat_lower` and
                        `yhat_upper`
                2. At most one of `ds` and `ts`

        Returns:
            A series where there is a probability corresponding to
                each instance (row) in the input.
        """
        residual = self._get_residual(y, yhat, yhat_lower, yhat_upper, residual)
        for _ in range(30):
            print(type(residual))

        log_proba = pd.DataFrame(
            {
                # pyre-fixme[16]: `KDEResidualTranslator` has no attribute `_kde`.
                "value": self._kde.score_samples(residual.value.to_frame()),
                "time": residual.time,
            },
            copy=False,
        )

        return TimeSeriesData(log_proba)

    def _get_residual(
        self,
        y: Optional[TimeSeriesData],
        yhat: Optional[TimeSeriesData],
        yhat_lower: Optional[TimeSeriesData],
        yhat_upper: Optional[TimeSeriesData],
        residual: Optional[TimeSeriesData],
    ) -> TimeSeriesData:
        if yhat is not None:
            if y is None:
                raise ValueError("Must include y if supplying yhat")
            if residual is not None:
                raise ValueError("Must not include residuals if supplying yhat")
            residual = y - yhat
            if (yhat_lower is not None) != (yhat_upper is not None):
                raise ValueError("Must supply either both yhat_lower and yhat_upper" "or neither")
            if yhat_lower is not None:
                assert yhat_upper is not None
                assert yhat_lower is not None
                residual /= yhat_upper - yhat_lower
        elif residual is not None:
            if any(c is not None for c in [y, yhat, yhat_lower, yhat_upper]):
                raise ValueError("Must not include y, yhat, yhat_lower, yhat_upper" "if supplying residuals")
        else:
            raise ValueError("Must supply y and yhat or residual")

        nulls = residual.value.isnull()
        residual.value = residual.value[~nulls]
        residual.time = residual.time[~nulls]
        return residual



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from unittest import TestCase

import numpy as np

from kats.compat import statsmodels
from kats.data.utils import load_air_passengers
from kats.detectors.trend_mk import MKDetector
from kats.models.theta import ThetaModel
from kats.models.theta import ThetaParams
from kats.utils.cupik import Pipeline


class cupikTest(TestCase):
    def setUp(self) -> None:
        self.TSData = load_air_passengers()

    def test_mkdetector(self) -> None:

        # We will be using 2 different scenarios to test if the results
        # are the same between a directly called MKDetector and one that
        # is called via CuPiK

        # Scene 1: window_size = 7, direction = 'up'
        pipe = Pipeline(
            # pyre-fixme[6]: For 1st param expected `List[Tuple[str, Step]]` but got
            #  `List[Tuple[str, MKDetector]]`.
            [
                ("trend_detector", MKDetector(threshold=0.8)),
            ]
        )
        pipe.fit(
            self.TSData,
            params={"trend_detector": {"window_size": 7, "direction": "up"}},
        )

        self.assertEqual(len(pipe.metadata["trend_detector"][0]), 50)
        self.assertEqual(
            len(pipe.metadata["trend_detector"][0]),
            len(MKDetector(data=self.TSData).detector(window_size=7, direction="up")),
        )

        # Scene 2: Default parameters of MKDetector
        pipe = Pipeline(
            # pyre-fixme[6]: For 1st param expected `List[Tuple[str, Step]]` but got
            #  `List[Tuple[str, MKDetector]]`.
            [
                ("trend_detector", MKDetector(threshold=0.8)),
            ]
        )
        pipe.fit(self.TSData)

        self.assertEqual(len(pipe.metadata["trend_detector"][0]), 2)
        self.assertEqual(
            len(pipe.metadata["trend_detector"][0]),
            len(MKDetector(data=self.TSData).detector()),
        )

    def test_thetamodel(self) -> None:
        pipe = Pipeline(
            # pyre-fixme[6]: For 1st param expected `List[Tuple[str, Step]]` but got
            #  `List[Tuple[str, ThetaModel]]`.
            [("theta_model", ThetaModel(data=self.TSData, params=ThetaParams()))]
        )
        fitted = pipe.fit(self.TSData)
        bools = (
            # pyre-fixme[16]: Optional type has no attribute `values`.
            ThetaModel(self.TSData, ThetaParams()).fit().fitted_values.values
            # pyre-fixme[16]: Item `List` of
            #  `Union[List[kats.consts.TimeSeriesData], TimeSeriesData]` has no
            #  attribute `fitted_values`.
            == fitted.fitted_values.values
        )
        self.assertEqual(np.sum(bools), 144)
        old_statsmodels = statsmodels.version < "0.12"
        expected = 433.328591954023 if old_statsmodels else 433.1270492317991
        # pyre-fixme[16]: Item `List` of `Union[List[kats.consts.TimeSeriesData],
        #  TimeSeriesData]` has no attribute `predict`.
        self.assertEqual(expected, fitted.predict(1).fcst.values[0])

        # test if the model can be built on the output from the detector
        pipe = Pipeline(
            # pyre-fixme[6]: For 1st param expected `List[Tuple[str, Step]]` but got
            #  `List[Tuple[str, Union[MKDetector, ThetaModel]]]`.
            [
                ("trend_detector", MKDetector(threshold=0.8)),
                ("theta_model", ThetaModel(data=self.TSData, params=ThetaParams())),
            ]
        )
        fitted = pipe.fit(self.TSData)
        self.assertEqual(len(pipe.metadata["trend_detector"][0]), 2)
        expected = 433.328591954023 if old_statsmodels else 433.1270492317991
        # pyre-fixme[16]: Item `List` of `Union[List[kats.consts.TimeSeriesData],
        #  TimeSeriesData]` has no attribute `predict`.
        self.assertEqual(expected, fitted.predict(1).fcst.values[0])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""
CUSUM stands for cumulative sum, it is a changepoint detection algorithm.

In the Kats implementation, it has two main components:

  1. Locate the change point: The algorithm iteratively estimates the means
      before and after the change point and finds the change point
      maximizing/minimizing the cusum value until the change point has
      converged. The starting point for the change point is at the middle.

  2. Hypothesis testing: Conducting log likelihood ratio test where the null
      hypothesis has no change point with one mean and the alternative
      hypothesis has a change point with two means.

And here are a few things worth mentioning:

  * We assume there is only one increase/decrease change point;
  * We use Gaussian distribution as the underlying model to calculate the cusum
      value and conduct the hypothesis test;

Typical usage example:

>>> # Univariate CUSUM
>>> timeseries = TimeSeriesData(...)
>>> detector = CusumDetector(timeseries)
>>> #Run detector
>>> changepoints = detector.detector()
>>> # Plot the results
>>> detector.plot(changepoints)

The usage is the same for multivariate CUSUM except that the time series needs
to be multivariate and that the plotting functions are not yet supported for
this use case.
"""

import logging
from dataclasses import asdict
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import chi2  # @manual

from kats.consts import TimeSeriesChangePoint
from kats.consts import TimeSeriesData
from kats.detectors.detector import Detector

pd.options.plotting.matplotlib.register_converters = True

_log: logging.Logger = logging.getLogger("cusum_detection")


@dataclass
class CUSUMDefaultArgs:
    threshold: float = 0.01
    max_iter: int = 10
    delta_std_ratio: float = 1.0
    min_abs_change: int = 0
    start_point: Optional[int] = None
    change_directions: Optional[List[str]] = None
    interest_window: Optional[int] = None
    magnitude_quantile: Optional[float] = None
    magnitude_ratio: float = 1.3
    magnitude_comparable_day: float = 0.5
    return_all_changepoints: bool = False
    remove_seasonality: bool = False


@dataclass
class CUSUMChangePointVal:
    changepoint: int
    mu0: float
    mu1: float
    changetime: List[float]
    stable_changepoint: bool
    delta: float
    llr_int: float
    p_value_int: float
    delta_int: Optional[float]
    sigma0: Optional[float] = None
    sigma1: Optional[float] = None
    llr: Optional[float] = None
    p_value: Optional[float] = None
    regression_detected: Optional[bool] = None


@dataclass
class VectorizedCUSUMChangePointVal:
    changepoint: List[int]
    mu0: List[float]
    mu1: List[float]
    changetime: List[float]
    stable_changepoint: List[bool]
    delta: List[float]
    llr_int: List[float]
    p_value_int: List[float]
    delta_int: Optional[List[float]]
    sigma0: Optional[List[float]] = None
    sigma1: Optional[List[float]] = None
    llr: Optional[List[float]] = None
    p_value: Optional[List[float]] = None
    regression_detected: Optional[List[bool]] = None


class CUSUMChangePoint(TimeSeriesChangePoint):
    """CUSUM change point.

    This is a changepoint detected by CUSUMDetector.

    Attributes:

        start_time: Start time of the change.
        end_time: End time of the change.
        confidence: The confidence of the change point.
        direction: a str stand for the changepoint change direction 'increase'
            or 'decrease'.
        cp_index: an int for changepoint index.
        mu0: a float indicates the mean before changepoint.
        mu1: a float indicates the mean after changepoint.
        delta: mu1 - mu0.
        llr: log likelihood ratio.
        llr_int: log likelihood ratio in the interest window.
        regression_detected: a bool indicates if regression detected.
        stable_changepoint: a bool indicates if we have a stable changepoint
            when locating the changepoint.
        p_value: p_value of the changepoint.
        p_value_int: p_value of the changepoint in the interest window.
    """

    def __init__(
        self,
        start_time: pd.Timestamp,
        end_time: pd.Timestamp,
        confidence: float,
        direction: str,
        cp_index: int,
        mu0: Union[float, np.ndarray],
        mu1: Union[float, np.ndarray],
        delta: Union[float, np.ndarray],
        llr_int: float,
        llr: float,
        regression_detected: bool,
        stable_changepoint: bool,
        p_value: float,
        p_value_int: float,
    ) -> None:
        super().__init__(start_time, end_time, confidence)
        self._direction = direction
        self._cp_index = cp_index
        self._mu0 = mu0
        self._mu1 = mu1
        self._delta = delta
        self._llr_int = llr_int
        self._llr = llr
        self._regression_detected = regression_detected
        self._stable_changepoint = stable_changepoint
        self._p_value = p_value
        self._p_value_int = p_value_int

    @property
    def direction(self) -> str:
        return self._direction

    @property
    def cp_index(self) -> int:
        return self._cp_index

    @property
    def mu0(self) -> Union[float, np.ndarray]:
        return self._mu0

    @property
    def mu1(self) -> Union[float, np.ndarray]:
        return self._mu1

    @property
    def delta(self) -> Union[float, np.ndarray]:
        return self._delta

    @property
    def llr(self) -> float:
        return self._llr

    @property
    def llr_int(self) -> float:
        return self._llr_int

    @property
    def regression_detected(self) -> bool:
        return self._regression_detected

    @property
    def stable_changepoint(self) -> bool:
        return self._stable_changepoint

    @property
    def p_value(self) -> float:
        return self._p_value

    @property
    def p_value_int(self) -> float:
        return self._p_value_int

    def __repr__(self) -> str:
        return (
            f"CUSUMChangePoint(start_time: {self._start_time}, end_time: "
            f"{self._end_time}, confidence: {self._confidence}, direction: "
            f"{self._direction}, index: {self._cp_index}, delta: {self._delta}, "
            f"regression_detected: {self._regression_detected}, "
            f"stable_changepoint: {self._stable_changepoint}, mu0: {self._mu0}, "
            f"mu1: {self._mu1}, llr: {self._llr}, llr_int: {self._llr_int}, "
            f"p_value: {self._p_value}, p_value_int: {self._p_value_int})"
        )


class CUSUMDetector(Detector):
    interest_window: Optional[Tuple[int, int]] = None
    magnitude_quantile: Optional[float] = None
    magnitude_ratio: Optional[float] = None
    changes_meta: Optional[Dict[str, Dict[str, Any]]] = None

    def __init__(
        self,
        data: TimeSeriesData,
        is_multivariate: bool = False,
        is_vectorized: bool = False,
    ) -> None:
        """Univariate CUSUM detector for level shifts

        Use cusum to detect changes, the algorithm is based on likelihood ratio
        cusum. See https://www.fs.isy.liu.se/Edu/Courses/TSFS06/PDFs/Basseville.pdf
        for details. This detector is used to detect mean changes in Normal
        Distribution.

        Args:

            data: :class:`kats.consts.TimeSeriesData`; The input time series data.
            is_multivariate: Optional; bool; should be False unless running
                MultiCUSUMDetector,
        """
        super(CUSUMDetector, self).__init__(data=data)
        if not self.data.is_univariate() and not is_multivariate and not is_vectorized:
            msg = (
                "CUSUMDetector only supports univariate time series, but got "
                f"{type(self.data.value)}.  For multivariate time series, use "
                "MultiCUSUMDetector or VectorizedCUSUMDetector"
            )
            _log.error(msg)
            raise ValueError(msg)

    def _get_change_point(
        self, ts: np.ndarray, max_iter: int, start_point: int, change_direction: str
    ) -> CUSUMChangePointVal:
        """
        Find change point in the timeseries.
        """
        interest_window = self.interest_window

        # locate the change point using cusum method
        if change_direction == "increase":
            changepoint_func = np.argmin
            _log.debug("Detecting increase changepoint.")
        else:
            assert change_direction == "decrease"
            changepoint_func = np.argmax
            _log.debug("Detecting decrease changepoint.")
        n = 0
        # use the middle point as initial change point to estimate mu0 and mu1
        if interest_window is not None:
            ts_int = ts[interest_window[0] : interest_window[1]]
        else:
            ts_int = ts

        if start_point is None:
            cusum_ts = np.cumsum(ts_int - np.mean(ts_int))
            changepoint = min(changepoint_func(cusum_ts), len(ts_int) - 2)
        else:
            changepoint = start_point

        mu0 = mu1 = None
        # iterate until the changepoint converage
        while n < max_iter:
            n += 1
            mu0 = np.mean(ts_int[: (changepoint + 1)])
            mu1 = np.mean(ts_int[(changepoint + 1) :])
            mean = (mu0 + mu1) / 2
            # here is where cusum is happening
            cusum_ts = np.cumsum(ts_int - mean)
            next_changepoint = max(1, min(changepoint_func(cusum_ts), len(ts_int) - 2))
            if next_changepoint == changepoint:
                break
            changepoint = next_changepoint

        if n == max_iter:
            _log.info("Max iteration reached and no stable changepoint found.")
            stable_changepoint = False
        else:
            stable_changepoint = True

        # llr in interest window
        if interest_window is None:
            llr_int = np.inf
            pval_int = np.NaN
            delta_int = None
        else:
            llr_int = self._get_llr(ts_int, mu0, mu1, changepoint)
            pval_int = 1 - chi2.cdf(llr_int, 2)
            delta_int = mu1 - mu0
            changepoint += interest_window[0]

        # full time changepoint and mean
        mu0 = np.mean(ts[: (changepoint + 1)])
        mu1 = np.mean(ts[(changepoint + 1) :])

        return CUSUMChangePointVal(
            changepoint=changepoint,
            mu0=mu0,
            mu1=mu1,
            changetime=self.data.time[changepoint],
            stable_changepoint=stable_changepoint,
            delta=mu1 - mu0,
            llr_int=llr_int,
            p_value_int=pval_int,
            delta_int=delta_int,
        )

    def _get_llr(
        self,
        ts: np.ndarray,
        mu0: float,
        mu1: float,
        changepoint: int,
        sigma0: Optional[float] = None,
        sigma1: Optional[float] = None,
    ) -> float:
        """
        Calculate the log likelihood ratio
        """
        scale = np.sqrt(
            (np.sum((ts[: (changepoint + 1)] - mu0) ** 2) + np.sum((ts[(changepoint + 1) :] - mu1) ** 2))
            / (len(ts) - 2)
        )
        mu_tilde, sigma_tilde = np.mean(ts), np.std(ts)

        if scale == 0:
            scale = sigma_tilde * 0.01

        llr = -2 * (
            self._log_llr(ts[: (changepoint + 1)], mu_tilde, sigma_tilde, mu0, scale)
            + self._log_llr(ts[(changepoint + 1) :], mu_tilde, sigma_tilde, mu1, scale)
        )
        return llr

    def _log_llr(self, x: np.ndarray, mu0: float, sigma0: float, mu1: float, sigma1: float) -> float:
        """Helper function to calculate log likelihood ratio.

        This function calculate the log likelihood ratio of two Gaussian
        distribution log(l(0)/l(1)).

        Args:
            x: the data value.
            mu0: mean of model 0.
            sigma0: std of model 0.
            mu1: mean of model 1.
            sigma1: std of model 1.

        Returns:
            the value of log likelihood ratio.
        """

        return np.sum(np.log(sigma1 / sigma0) + 0.5 * (((x - mu1) / sigma1) ** 2 - ((x - mu0) / sigma0) ** 2))

    def _magnitude_compare(self, ts: np.ndarray) -> float:
        """
        Compare daily magnitude to avoid daily seasonality false positives.
        """
        time = self.data.time
        interest_window = self.interest_window
        magnitude_ratio = self.magnitude_ratio
        if interest_window is None:
            raise ValueError("detect must be called first")
        assert magnitude_ratio is not None

        # get number of days in historical window
        days = (time.max() - time.min()).days

        # get interest window magnitude
        mag_int = self._get_time_series_magnitude(ts[interest_window[0] : interest_window[1]])

        comparable_mag = 0

        for i in range(days):
            start_time = time[interest_window[0]] - pd.Timedelta(f"{i}D")
            end_time = time[interest_window[1]] - pd.Timedelta(f"{i}D")
            start_idx = time[time == start_time].index[0]
            end_idx = time[time == end_time].index[0]

            hist_int = self._get_time_series_magnitude(ts[start_idx:end_idx])
            if mag_int / hist_int >= magnitude_ratio:
                comparable_mag += 1

        return comparable_mag / days

    def _get_time_series_magnitude(self, ts: np.ndarray) -> float:
        """
        Calculate the magnitude of a time series.
        """
        magnitude = np.quantile(ts, self.magnitude_quantile, interpolation="nearest")
        return magnitude

    # pyre-fixme[14]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(self, **kwargs: Any) -> Sequence[CUSUMChangePoint]:
        """
        Find the change point and calculate related statistics.

        Args:

            threshold: Optional; float; significance level, default: 0.01.
            max_iter: Optional; int, maximum iteration in finding the
                changepoint.
            delta_std_ratio: Optional; float; the mean delta have to larger than
                this parameter times std of the data to be consider as a change.
            min_abs_change: Optional; int; minimal absolute delta between mu0
                and mu1.
            start_point: Optional; int; the start idx of the changepoint, if
                None means the middle of the time series.
            change_directions: Optional; list<str>; a list contain either or
                both 'increase' and 'decrease' to specify what type of change
                want to detect.
            interest_window: Optional; list<int, int>, a list containing the
                start and end of interest windows where we will look for change
                points. Note that llr will still be calculated using all data
                points.
            magnitude_quantile: Optional; float; the quantile for magnitude
                comparison, if none, will skip the magnitude comparison.
            magnitude_ratio: Optional; float; comparable ratio.
            magnitude_comparable_day: Optional; float; maximal percentage of
                days can have comparable magnitude to be considered as
                regression.
            return_all_changepoints: Optional; bool; return all the changepoints
                found, even the insignificant ones.

        Returns:
            A list of CUSUMChangePoint.
        """
        defaultArgs = CUSUMDefaultArgs()
        # Extract all arg values or assign defaults from default vals constant
        threshold = kwargs.get("threshold", defaultArgs.threshold)
        max_iter = kwargs.get("max_iter", defaultArgs.max_iter)
        delta_std_ratio = kwargs.get("delta_std_ratio", defaultArgs.delta_std_ratio)
        min_abs_change = kwargs.get("min_abs_change", defaultArgs.min_abs_change)
        start_point = kwargs.get("start_point", defaultArgs.start_point)
        change_directions = kwargs.get("change_directions", defaultArgs.change_directions)
        interest_window = kwargs.get("interest_window", defaultArgs.interest_window)
        magnitude_quantile = kwargs.get("magnitude_quantile", defaultArgs.magnitude_quantile)
        magnitude_ratio = kwargs.get("magnitude_ratio", defaultArgs.magnitude_ratio)
        magnitude_comparable_day = kwargs.get("magnitude_comparable_day", defaultArgs.magnitude_comparable_day)
        return_all_changepoints = kwargs.get("return_all_changepoints", defaultArgs.return_all_changepoints)

        self.interest_window = interest_window
        self.magnitude_quantile = magnitude_quantile
        self.magnitude_ratio = magnitude_ratio

        # Use array to store the data
        ts = self.data.value.to_numpy()
        ts = ts.astype("float64")
        changes_meta = {}

        if change_directions is None:
            change_directions = ["increase", "decrease"]

        for change_direction in change_directions:
            if change_direction not in {"increase", "decrease"}:
                raise ValueError("Change direction must be 'increase' or 'decrease.' " f"Got {change_direction}")

            change_meta = self._get_change_point(
                ts,
                max_iter=max_iter,
                start_point=start_point,
                change_direction=change_direction,
            )
            change_meta.llr = llr = self._get_llr(
                ts,
                change_meta.mu0,
                change_meta.mu1,
                change_meta.changepoint,
                change_meta.sigma0,
                change_meta.sigma1,
            )
            change_meta.p_value = 1 - chi2.cdf(llr, 2)

            # compare magnitude on interest_window and historical_window
            if np.min(ts) >= 0:
                if magnitude_quantile and interest_window:
                    change_ts = ts if change_direction == "increase" else -ts
                    mag_change = self._magnitude_compare(change_ts) >= magnitude_comparable_day
                else:
                    mag_change = True
            else:
                mag_change = True
                if magnitude_quantile:
                    _log.warning(("The minimal value is less than 0. Cannot perform " "magnitude comparison."))

            if_significant = llr > chi2.ppf(1 - threshold, 2)
            if_significant_int = change_meta.llr_int > chi2.ppf(1 - threshold, 2)
            if change_direction == "increase":
                larger_than_min_abs_change = change_meta.mu0 + min_abs_change < change_meta.mu1
            else:
                larger_than_min_abs_change = change_meta.mu0 > change_meta.mu1 + min_abs_change
            larger_than_std = np.abs(change_meta.delta) > np.std(ts[: change_meta.changepoint]) * delta_std_ratio

            change_meta.regression_detected = (
                if_significant and if_significant_int and larger_than_min_abs_change and larger_than_std and mag_change
            )
            changes_meta[change_direction] = asdict(change_meta)

        self.changes_meta = changes_meta

        return self._convert_cusum_changepoints(changes_meta, return_all_changepoints)

    def _convert_cusum_changepoints(
        self,
        cusum_changepoints: Dict[str, Dict[str, Any]],
        return_all_changepoints: bool,
    ) -> List[CUSUMChangePoint]:
        """
        Convert the output from the other kats cusum algorithm into
        CUSUMChangePoint type.
        """
        converted = []
        detected_cps = cusum_changepoints

        for direction in detected_cps:
            dir_cps = detected_cps[direction]
            if dir_cps["regression_detected"] or return_all_changepoints:
                # we have a change point
                change_point = CUSUMChangePoint(
                    start_time=dir_cps["changetime"],
                    end_time=dir_cps["changetime"],
                    confidence=1 - dir_cps["p_value"],
                    direction=direction,
                    cp_index=dir_cps["changepoint"],
                    mu0=dir_cps["mu0"],
                    mu1=dir_cps["mu1"],
                    delta=dir_cps["delta"],
                    llr_int=dir_cps["llr_int"],
                    llr=dir_cps["llr"],
                    regression_detected=dir_cps["regression_detected"],
                    stable_changepoint=dir_cps["stable_changepoint"],
                    p_value=dir_cps["p_value"],
                    p_value_int=dir_cps["p_value_int"],
                )
                converted.append(change_point)

        return converted

    def plot(self, change_points: Sequence[CUSUMChangePoint], **kwargs: Any) -> plt.Axes:
        """Plot detection results from CUSUM.

        Args:
            change_points: A list of CUSUMChangePoint.
            kwargs: other arguments to pass to subplots.

        Returns:
            The matplotlib Axes.
        """
        time_col_name = self.data.time.name
        val_col_name = self.data.value.name

        data_df = self.data.to_dataframe()

        _, ax = plt.subplots(**kwargs)
        ax.plot(data_df[time_col_name], data_df[val_col_name])

        changepoint_annotated = False
        for change in change_points:
            if change.regression_detected:
                ax.axvline(x=change.start_time, color="red")
                changepoint_annotated = True
        if not changepoint_annotated:
            _log.warning("No change points detected!")

        interest_window = self.interest_window
        if interest_window is not None:
            ax.axvspan(
                pd.to_datetime(self.data.time)[interest_window[0]],
                pd.to_datetime(self.data.time)[interest_window[1] - 1],
                alpha=0.3,
                label="interets_window",
            )

        return ax


class MultiCUSUMDetector(CUSUMDetector):
    """
    MultiCUSUM is similar to univariate CUSUM, but we use MultiCUSUM to find a
    changepoint in multivariate time series.  The detector is used to detect
    changepoints in the multivariate mean of the time series.  The cusum values
    and likelihood ratio test calculations assume the underlying distribution
    has a Multivariate Guassian distriubtion.

    Attributes:
        data: The input time series data from TimeSeriesData
    """

    def __init__(self, data: TimeSeriesData) -> None:
        super(MultiCUSUMDetector, self).__init__(data=data, is_multivariate=True)

    def detector(self, **kwargs: Any) -> List[CUSUMChangePoint]:
        """
        Overwrite the detector method for MultiCUSUMDetector.

        Args:
            threshold: Optional; float; significance level, default: 0.01.
            max_iter: Optional; int, maximum iteration in finding the
                changepoint.
            start_point: Optional; int; the start idx of the changepoint, if
                None means the middle of the time series.
        """
        defaultArgs = CUSUMDefaultArgs()
        # Extract all arg values or assign defaults from default vals constant
        threshold = kwargs.get("threshold", defaultArgs.threshold)
        max_iter = kwargs.get("max_iter", defaultArgs.max_iter)
        start_point = kwargs.get("start_point", defaultArgs.start_point)

        # TODO: Add support for interest windows

        return_all_changepoints = kwargs.get("return_all_changepoints", defaultArgs.return_all_changepoints)

        # Use array to store the data
        ts = self.data.value.to_numpy()
        ts = ts.astype("float64")
        changes_meta = {}

        # We will always be looking for increases in the CUSUM values for
        # multivariate detection. We keep using change_direction = "increase"
        # here to be consistent with the univariate detector.
        for change_direction in ["increase"]:

            change_meta = self._get_change_point(
                ts,
                max_iter=max_iter,
                start_point=start_point,
            )
            change_meta.llr = llr = self._get_llr(
                ts,
                change_meta.mu0,
                change_meta.mu1,
                change_meta.changepoint,
                change_meta.sigma0,
                change_meta.sigma1,
            )
            change_meta.p_value = 1 - chi2.cdf(llr, ts.shape[1] + 1)

            if_significant = llr > chi2.ppf(1 - threshold, ts.shape[1] + 1)

            change_meta.regression_detected = if_significant
            changes_meta[change_direction] = asdict(change_meta)

        self.changes_meta = changes_meta

        return self._convert_cusum_changepoints(changes_meta, return_all_changepoints)

    def _get_llr(
        self,
        ts: np.ndarray,
        mu0: float,
        mu1: float,
        changepoint: int,
        sigma0: Optional[float],
        sigma1: Optional[float],
    ) -> float:

        mu_tilde = np.mean(ts, axis=0)
        sigma_pooled = np.cov(ts, rowvar=False)
        llr = -2 * (
            self._log_llr_multi(
                ts[: (changepoint + 1)],
                mu_tilde,
                sigma_pooled,
                mu0,
                sigma0,  # pyre-fixme
            )
            - self._log_llr_multi(
                ts[(changepoint + 1) :],
                mu_tilde,
                sigma_pooled,
                mu1,
                sigma1,  # pyre-fixme
            )
        )
        return llr

    def _log_llr_multi(
        self,
        x: np.ndarray,
        mu0: Union[float, np.ndarray],
        sigma0: Union[float, np.ndarray],
        mu1: Union[float, np.ndarray],
        sigma1: Union[float, np.ndarray],
    ) -> float:
        try:
            sigma0_inverse = np.linalg.inv(sigma0)
            sigma1_inverse = np.linalg.inv(sigma1)
            log_det_sigma0 = np.log(np.linalg.det(sigma0))
            log_det_sigma1 = np.log(np.linalg.det(sigma1))
        except np.linalg.linalg.LinAlgError:
            msg = "One or more covariance matrix is singular."
            _log.error(msg)
            raise ValueError(msg)

        return len(x) / 2 * (log_det_sigma0 - log_det_sigma1) + np.sum(
            -np.matmul(np.matmul(x[i] - mu1, sigma1_inverse), (x[i] - mu1).T)
            + np.matmul(np.matmul(x[i] - mu0, sigma0_inverse), (x[i] - mu0).T)
            for i in range(len(x))
        )

    def _get_change_point(
        self,
        ts: np.ndarray,
        max_iter: int,
        start_point: int,
        change_direction: str = "increase",
    ) -> CUSUMChangePointVal:

        # locate the change point using cusum method
        changepoint_func = np.argmin
        n = 0
        ts_int = ts

        if start_point is None:
            start_point = len(ts_int) // 2
        changepoint = start_point

        # iterate until the changepoint converage
        while n < max_iter:
            n += 1
            data_before_changepoint = ts_int[: (changepoint + 1)]
            data_after_changepoint = ts_int[(changepoint + 1) :]

            mu0 = np.mean(data_before_changepoint, axis=0)
            mu1 = np.mean(data_after_changepoint, axis=0)

            # TODO: replace pooled variance with sample variances before and
            # after changepoint.
            # sigma0 = np.cov(data_before_changepoint, rowvar=False)
            # sigma1 = np.cov(data_after_changepoint, rowvar=False)
            sigma0 = sigma1 = np.cov(ts_int, rowvar=False)

            try:
                log_det_sigma0 = np.log(np.linalg.det(sigma0))
                log_det_sigma1 = np.log(np.linalg.det(sigma1))
                sigma0_inverse = np.linalg.inv(sigma0)
                sigma1_inverse = np.linalg.inv(sigma1)
            except np.linalg.linalg.LinAlgError:
                msg = "One or more covariance matrix is singular."
                _log.error(msg)
                raise ValueError(msg)

            si_values = np.diag(
                -(1 / 2) * log_det_sigma1
                - np.matmul(np.matmul(ts_int - mu1, sigma1_inverse), (ts_int - mu1).T)
                + (1 / 2) * log_det_sigma0
                + np.matmul(np.matmul(ts_int - mu0, sigma0_inverse), (ts_int - mu0).T)
            )

            cusum_ts = np.cumsum(si_values)
            next_changepoint = max(1, min(changepoint_func(cusum_ts), len(cusum_ts) - 2))

            if next_changepoint == changepoint:
                break
            else:
                changepoint = next_changepoint

        if n == max_iter:
            _log.info("Max iteration reached and no stable changepoint found.")
            stable_changepoint = False
        else:
            stable_changepoint = True

        llr_int = np.inf
        pval_int = np.NaN
        delta_int = None

        # full time changepoint and mean
        mu0 = np.mean(ts[: (changepoint + 1)], axis=0)
        mu1 = np.mean(ts[(changepoint + 1) :], axis=0)
        sigma0 = sigma1 = np.cov(ts, rowvar=False)

        return CUSUMChangePointVal(
            changepoint=changepoint,
            mu0=mu0,
            mu1=mu1,
            changetime=self.data.time[changepoint],
            stable_changepoint=stable_changepoint,
            delta=mu1 - mu0,
            llr_int=llr_int,
            p_value_int=pval_int,
            delta_int=delta_int,
            sigma0=sigma0,
            sigma1=sigma1,
        )


class VectorizedCUSUMDetector(CUSUMDetector):
    """
    VectorizedCUSUM is the vecteorized version of CUSUM. It can take
    multiple time series as an input and run CUSUM algorithm on each time series
    in a vectorized manner.

    Attributes:
        data: The input time series data from TimeSeriesData
    """

    changes_meta_list: Optional[List[Dict[str, Dict[str, Any]]]] = None

    def __init__(self, data: TimeSeriesData) -> None:
        super(VectorizedCUSUMDetector, self).__init__(data=data, is_multivariate=False, is_vectorized=True)

    def detector(self, **kwargs: Any) -> Sequence[CUSUMChangePoint]:
        msg = "VectorizedCUSUMDetector is in beta and please use detector_()"
        _log.error(msg)
        raise ValueError(msg)

    def detector_(self, **kwargs: Any) -> List[List[CUSUMChangePoint]]:
        """
        Detector method for vectorized version of CUSUM

        Args:
            threshold: Optional; float; significance level, default: 0.01.
            max_iter: Optional; int, maximum iteration in finding the
                changepoint.
            delta_std_ratio: Optional; float; the mean delta have to larger than
                this parameter times std of the data to be consider as a change.
            min_abs_change: Optional; int; minimal absolute delta between mu0
                and mu1.
            change_directions: Optional; list<str>; a list contain either or
                both 'increase' and 'decrease' to specify what type of change
                want to detect.
            interest_window: Optional; list<int, int>, a list containing the
                start and end of interest windows where we will look for change
                points. Note that llr will still be calculated using all data
                points.
            magnitude_quantile: Optional; float; the quantile for magnitude
                comparison, if none, will skip the magnitude comparison.
            magnitude_ratio: Optional; float; comparable ratio.
            magnitude_comparable_day: Optional; float; maximal percentage of
                days can have comparable magnitude to be considered as
                regression.
            return_all_changepoints: Optional; bool; return all the changepoints
                found, even the insignificant ones.

        Returns:
            A list of tuple of TimeSeriesChangePoint and CUSUMMetadata.
        """
        defaultArgs = CUSUMDefaultArgs()
        # Extract all arg values or assign defaults from default vals constant
        threshold = kwargs.get("threshold", defaultArgs.threshold)
        max_iter = kwargs.get("max_iter", defaultArgs.max_iter)
        delta_std_ratio = kwargs.get("delta_std_ratio", defaultArgs.delta_std_ratio)
        min_abs_change = kwargs.get("min_abs_change", defaultArgs.min_abs_change)
        change_directions = kwargs.get("change_directions", defaultArgs.change_directions)
        interest_window = kwargs.get("interest_window", defaultArgs.interest_window)
        magnitude_quantile = kwargs.get("magnitude_quantile", defaultArgs.magnitude_quantile)
        magnitude_ratio = kwargs.get("magnitude_ratio", defaultArgs.magnitude_ratio)
        magnitude_comparable_day = kwargs.get("magnitude_comparable_day", defaultArgs.magnitude_comparable_day)
        return_all_changepoints = kwargs.get("return_all_changepoints", defaultArgs.return_all_changepoints)
        self.interest_window = interest_window
        self.magnitude_quantile = magnitude_quantile
        self.magnitude_ratio = magnitude_ratio

        # Use array to store the data
        ts_all = self.data.value.to_numpy()
        ts_all = ts_all.astype("float64")
        if ts_all.ndim == 1:
            ts_all = ts_all[:, np.newaxis]
        changes_meta_list = []

        if change_directions is None:
            change_directions = ["increase", "decrease"]

        change_meta_all = {}
        for change_direction in change_directions:
            if change_direction not in {"increase", "decrease"}:
                raise ValueError("Change direction must be 'increase' or 'decrease.' " f"Got {change_direction}")

            change_meta_all[change_direction] = asdict(
                self._get_change_point_multiple_ts(
                    ts_all,
                    max_iter=max_iter,
                    change_direction=change_direction,
                )
            )

        ret = []
        for col_idx in np.arange(ts_all.shape[1]):
            ts = ts_all[:, col_idx]
            changes_meta = {}
            for change_direction in change_directions:
                change_meta_ = change_meta_all[change_direction]
                # if no change points are detected, skip
                if not list(change_meta_["changepoint"]):
                    continue
                change_meta = {
                    k: change_meta_[k][col_idx]
                    if isinstance(change_meta_[k], np.ndarray) or isinstance(change_meta_[k], list)
                    else change_meta_[k]
                    for k in change_meta_
                }
                change_meta["llr"] = llr = self._get_llr(
                    ts,
                    change_meta["mu0"],
                    change_meta["mu1"],
                    change_meta["changepoint"],
                    change_meta["sigma0"],
                    change_meta["sigma1"],
                )
                change_meta["p_value"] = 1 - chi2.cdf(change_meta["llr"], 2)

                # compare magnitude on interest_window and historical_window
                if np.min(ts) >= 0:
                    if magnitude_quantile and interest_window:
                        change_ts = ts if change_direction == "increase" else -ts
                        mag_change = self._magnitude_compare(change_ts) >= magnitude_comparable_day
                    else:
                        mag_change = True
                else:
                    mag_change = True
                    if magnitude_quantile:
                        _log.warning(("The minimal value is less than 0. Cannot perform " "magnitude comparison."))

                if_significant = llr > chi2.ppf(1 - threshold, 2)
                if_significant_int = change_meta["llr_int"] > chi2.ppf(1 - threshold, 2)
                if change_direction == "increase":
                    larger_than_min_abs_change = change_meta["mu0"] + min_abs_change < change_meta["mu1"]
                else:
                    larger_than_min_abs_change = change_meta["mu0"] > change_meta["mu1"] + min_abs_change
                larger_than_std = (
                    np.abs(change_meta["delta"]) > np.std(ts[: change_meta["changepoint"]]) * delta_std_ratio
                )

                change_meta["regression_detected"] = (
                    if_significant
                    and if_significant_int
                    and larger_than_min_abs_change
                    and larger_than_std
                    and mag_change
                )

                changes_meta[change_direction] = change_meta
            changes_meta_list.append(changes_meta)
            ret.append(self._convert_cusum_changepoints(changes_meta, return_all_changepoints))
        self.changes_meta_list = changes_meta_list
        return ret

    def _get_change_point_multiple_ts(
        self, ts: np.ndarray, max_iter: int, change_direction: str
    ) -> VectorizedCUSUMChangePointVal:
        """
        Find change points in a list of time series
        """
        interest_window = self.interest_window

        # locate the change point using cusum method
        if change_direction == "increase":
            changepoint_func = np.argmin
            _log.debug("Detecting increase changepoint.")
        else:
            assert change_direction == "decrease"
            changepoint_func = np.argmax
            _log.debug("Detecting decrease changepoint.")

        n = 0
        if interest_window is not None:
            ts_int = ts[interest_window[0] : interest_window[1], :]
        else:
            ts_int = ts
        n_ts = ts_int.shape[1]
        n_pts = ts_int.shape[0]
        if n_pts == 0:
            return VectorizedCUSUMChangePointVal(
                changepoint=[],
                mu0=[],
                mu1=[],
                changetime=[],
                stable_changepoint=[],
                delta=[],
                llr_int=[],
                p_value_int=[],
                delta_int=[],
            )
        # use the middle point as initial change point to estimate mu0 and mu1
        tmp = ts_int - np.tile(np.mean(ts_int, axis=0), (n_pts, 1))
        cusum_ts = np.cumsum(tmp, axis=0)
        changepoint = np.minimum(changepoint_func(cusum_ts, axis=0), n_pts - 2)
        mu0 = mu1 = None
        stable_changepoint = [False] * len(changepoint)
        # iterate until the changepoint converage
        while n < max_iter:
            mask = np.zeros((n_pts, n_ts), dtype=bool)
            for i, c in enumerate(changepoint):
                mask[: (c + 1), i] = True
            n += 1
            mu0 = np.divide(np.sum(np.multiply(ts_int, mask), axis=0), np.sum(mask, axis=0))
            mu1 = np.divide(np.sum(np.multiply(ts_int, ~mask), axis=0), np.sum(~mask, axis=0))
            mean = (mu0 + mu1) / 2
            # here is where cusum is happening
            tmp = ts_int - np.tile(mean, (n_pts, 1))
            cusum_ts = np.cumsum(tmp, axis=0)
            next_changepoint = np.maximum(np.minimum(changepoint_func(cusum_ts, axis=0), n_pts - 2), 1)
            stable_changepoint = np.equal(changepoint, next_changepoint)
            if all(stable_changepoint):
                break
            changepoint = next_changepoint

        # full time changepoint and mean
        mask = np.zeros((n_pts, n_ts), dtype=bool)
        for i, c in enumerate(changepoint):
            mask[: (c + 1), i] = True
        mu0 = np.divide(np.sum(np.multiply(ts_int, mask), axis=0), np.sum(mask, axis=0))
        mu1 = np.divide(np.sum(np.multiply(ts_int, ~mask), axis=0), np.sum(~mask, axis=0))

        # llr in interest window
        if interest_window is None:
            llr_int = [np.inf for _ in np.arange(n_ts)]
            pval_int = [np.NaN for _ in np.arange(n_ts)]
            delta_int = None
        else:
            llr_int = []
            pval_int = []
            delta_int = []
            for col_idx in np.arange(n_ts):
                _llr_int = self._get_llr(ts_int[:, col_idx], mu0[col_idx], mu1[col_idx], changepoint[col_idx])
                _pval_int = 1 - chi2.cdf(_llr_int, 2)
                _delta_int = mu1[col_idx] - mu0[col_idx]
                llr_int.append(_llr_int)
                pval_int.append(_pval_int)
                delta_int.append(_delta_int)
                changepoint[col_idx] += interest_window[0]

        return VectorizedCUSUMChangePointVal(
            changepoint=changepoint,
            mu0=mu0,
            mu1=mu1,
            changetime=[self.data.time[c] for c in changepoint],
            stable_changepoint=stable_changepoint,
            delta=mu1 - mu0,
            llr_int=llr_int,
            p_value_int=pval_int,
            delta_int=delta_int,
        )
